# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
This module contains useful function to extract data from nodes,
specific to aiida-fleur. Any useful code snipets for handling large number of
nodes, or data-mine go in here
"""

#import time
from aiida.orm import DataFactory, Node, load_node
from aiida.orm.querybuilder import QueryBuilder as QB


StructureData = DataFactory('structure')
ParameterData = DataFactory('parameter')




def extract_structure_info(keys, structures=None):
    """
    A method that collects a bunch of information (specified in keys) from 
    structures (default whole db, or provided node list) in the database and 
    returns that information as a dict, which could be used for further evalation
    #keys = ['uuid', 'formula', 'pk', 'symmetry', 'pbc', 'volume', 'total_energy', 
    'child_nodes', 'natoms', 'group', extras', 'label', 'description', 'cif_file', 
    'cif_number', 'cif_uuid', 'cif_ref', 'workfunctions', 'band', 'dos', 'eos', 
    'init_cls', 'corehole', primitive]

    """
    structure_list = []
    
    from aiida_fleur.tools.StructureData_util import get_spacegroup, is_structure
    from aiida_fleur.tools.StructureData_util import is_primitive

    if not structures: 
        StructureData = DataFactory('structure')
        #t = time.time()
        qb = QB()
        qb.append(StructureData)
        structures = qb.all()  
        #elapsed = time.time() - t
        #print "Total number of structures: {} (retrieved in {} s.)".format(len(structures), elapsed)
        #t = time.time()
    
    #for structure in structures:
    #    structure_dict = {}
    #    struc = structure[0]
    #    for key in keys:
    #        structure_dict[key] = get_methods(key)(struc)
    
    #get information
    for structure in structures:
        structure_dict = {}
        
        if isinstance(structure, list):
            struc = structure[0]
        else:
            struc = is_structure(structure)
        
        if 'formula' in keys:
            structure_dict['formula'] = struc.get_formula()
        if 'pk' in keys:
            structure_dict['pk'] = struc.pk
        if 'uuid' in keys:
            structure_dict['uuid'] = str(struc.uuid)
        if 'natoms' in keys:
            structure_dict['natoms'] = len(struc.sites)
        if 'cell' in keys:
            structure_dict['cell'] = str(struc.cell)
        if 'pbc' in keys:
            structure_dict['pbc'] = str(struc.pbc)
        if 'label' in keys:
            structure_dict['label'] = struc.label
        if 'description' in keys:
            structure_dict['description'] = struc.description
        if 'extras' in keys:
            extras = struc.get_extras()
            structure_dict['extras'] = str(extras)
        if 'symmetry' in keys:
            symmetry = get_spacegroup(struc)
            structure_dict['symmetry'] = str(symmetry)
        if 'volume' in keys:
            volume = struc.get_cell_volume()
            structure_dict['volume'] = volume
        if 'child_nodes' in keys:
            child_nodes = len(struc.get_outputs())
            structure_dict['child_nodes'] = child_nodes
        if 'primitive' in keys:
            prim = is_primitive(struc)
            structure_dict['primitive'] = prim
        
        if 'cif_file' in keys:
            cif_file = get_cif_file(struc)
            structure_dict['cif_file'] = cif_file
        '''
        if 'cif_number' in keys:
            cif_number = get_cif_number(struc)
            structure_dict['cif_number'] = cif_number
        if 'cif_uuid' in keys:
            cif_uuid = get_cif_uuid(struc)
            structure_dict['cif_uuid'] = cif_uuid
        if 'cif_ref' in keys:
            cif_ref = get_cif_ref(struc)
            structure_dict['cif_ref'] = cif_ref
        if 'total_energy' in keys:
            total_energy = get_total_energy(struc)
            structure_dict['total_energy'] = total_energy
        '''
        if 'group' in keys:
            group = group_member(struc)
            structure_dict['group'] = group         
        if 'scf' in keys:
            scf = input_of_workcal('fleur_scf_wc', struc)
            structure_dict['scf'] = scf        
        if 'band' in keys:
            band = input_of_workcal('fleur_band_wc', struc)
            structure_dict['band'] = band    
        if 'dos' in keys:
            dos = input_of_workcal('fleur_dos_wc', struc)
            structure_dict['dos'] = dos
        if 'eos' in keys:
            eos = input_of_workcal('fleur_eos_wc', struc)
            structure_dict['eos'] = eos
        if 'init_cls' in keys:
            init_cls = input_of_workcal('fleur_initial_cls_wc', struc)
            structure_dict['init_cls'] = init_cls
        if 'corehole' in keys:
            corehole = input_of_workcal('fleur_corehole_wc', struc)
            structure_dict['corehole'] = corehole 
        if 'workfunctions' in keys:
            workfunctions_uuid, workfunctions_name = input_of_workfunctions(struc)
            structure_dict['workfunctions'] = [workfunctions_uuid, workfunctions_name] 
            
        structure_list.append(structure_dict)
        
    #elapsed = time.time() - t
    #print "(needed {} s.!!!)".format(elapsed)
    
    return structure_list
    

def group_member(node):
    """
    Find to what groups a node belongs to.
    
    Comment: currently very greedy!
    """
    from aiida.orm.group import Group
    member_in = []
    #get all groups in db
    # for each group check if node is member of group
    # append group name to member_in

    res = Group.query()
    for group in res:
        name = group.name
        for gnode in group.nodes: # TODO: easier/better way?
            if gnode.uuid == node.uuid:
                member_in.append(name)
    return member_in

def input_of_workcal(name, node):
    """
    checks if a given node was input into a certain workcalculation
    and returns a list of workcalculation uuids of workcalculations with the given name
    """
    from aiida.orm.implementation.general.calculation.work import WorkCalculation
    process_uuids = []
    for out in node.get_outputs():
        if isinstance(out, WorkCalculation):
            label = out.get_attr('_process_label')
            if label == name:
                process_uuids.append(out.uuid)
    return process_uuids

def input_of_workfunctions(node, name=''):
    """
    checks if a given node was input into a certain workfunction
    and returns a list of workfunction uuids of workfunction with the given name
    """
    from aiida.orm.implementation.general.calculation.work import WorkCalculation
    process_uuids = []
    process_names = []
    outputs = node.get_outputs()
    for out in outputs:
        if isinstance(out, WorkCalculation):
            try:# TODO: is there a better way
                label = out.get_attr('function_name')
            except AttributeError:
                label = None
                continue
            if label == name:
                process_uuids.append(out.uuid)
                process_names.append(label)
            if not name:
                process_uuids.append(out.uuid)
                process_names.append(label)
    return process_uuids, process_names

def get_cif_file(node):
    """
    Finds out if (a structure) given as input was created from a cif file
    currently with the method wf_struc_from_cif
    
    params: node: structureData node
    returns [cif_filename, cif_uuid, cif_folder]
    """
    from aiida.orm.implementation.general.calculation.work import WorkCalculation
    from aiida.orm.data.cif import CifData
    inputs = node.get_inputs()
    name = 'wf_struc_from_cif'# TODO: Bad solution, not general for me currently enough
    cif_uuid, cif_filename, cif_folder = '','',''
    for inp in inputs:
        if isinstance(inp, WorkCalculation):
            try:# TODO: is there a better way
                label = inp.get_attr('function_name') #process_label
            except AttributeError:
                label = None
                continue
            if label == name:
                inp_wc = inp.get_inputs()
                for iwc in inp_wc:
                    if isinstance(iwc, CifData):
                        cif_uuid = iwc.uuid
                        cif_filename = iwc.filename
                        cif_folder = iwc.folder.abspath
                        break
    
    return [cif_filename, cif_uuid, cif_folder]
#wf_struc_from_cif process_label
#cif.filename
#cif.uuid
#cif.folder.abspath
#wc.get_inputs()


# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
This module contains the parser for a FLEUR calculation and methods for parsing
different files produced by FLEUR.

Please implement file parsing routines that they can be executed from outside
the parser. Makes testing and portability easier.
"""
# TODO: cleanup
# TODO: move methods to utils, xml or other
# TODO: warnings
import os
#import numpy
from datetime import date
from aiida.orm.data.parameter import ParameterData
from aiida.parsers.parser import Parser
from aiida.orm.data.array.bands import BandsData
from aiida_fleur.calculation.fleur import FleurCalculation
from aiida_fleur.parsers import FleurOutputParsingError
from aiida_fleur.data.fleurinp import FleurinpData


class FleurParser(Parser):
    """
    This class is the implementation of the Parser class for FLEUR.
    It parses the FLEUR output if the calculation was successful,
    i.e checks if all files are there that should be and their condition.
    Then it parses the out.xml file and returns a (simple) parameterData node
    with the results of the last iteration.
    Other files (DOS.x, bands.x, ...) are also parsed if they are retrieved.
    """

    _setting_key = 'parser_options'

    def __init__(self, calc):
        """
        Initialize the instance of FleurParser
        """
        # check for valid input
        if not isinstance(calc, FleurCalculation):
            raise FleurOutputParsingError("Input calculation for the FleurParser"
                                          "must be a FleurCalculation")

        # these files should be at least present after success of a Fleur run

        self._default_files = {calc._OUTXML_FILE_NAME, calc._INPXML_FILE_NAME}
        self._other_files = {}#"enpara","inp","sym.out", "cdn1", }
        #plus other special files? corelevels.xx, DOS.1 Bands.1 ...
        self._should_retrieve = []#calc.Calcinfo.retrieve_list()
        # somehow calc.CalcInfo does not work..
        super(FleurParser, self).__init__(calc)

    def get_linkname_outparams_complex(self):
        """
        Returns the name of the link to the output_complex
        Node contains the Fleur output in a rather complex dictionary.
        """
        return 'output_complex'


    def parse_with_retrieved(self, retrieved):
        """
        Receives in input a dictionary of retrieved nodes.
        Does all the logic here. Checks presents of files.
        Calls routines to parse them and returns parameter nodes and sucess.

        :return successful: Bool, if overall parsing was successful or not
        :return new_nodes_list: list of tuples of two (linkname, Dataobject),
                                nodes to be stored by AiiDA

        """

        ####### init some variables ######

        successful = True

        has_xml_outfile = False
        has_dos_file = False
        has_bands_file = False
        has_new_xmlinp_file = False

        dos_file = None
        band_file = None
        new_nodes_list = []

        ######### Check presents of files ######

        # select the folder object
        # Check that the retrieved folder is there
        try:
            out_folder = retrieved[self._calc._get_linkname_retrieved()]
        except KeyError:
            self.logger.error(u"No retrieved folder found")
            return False, ()

        # check what is inside the folder
        list_of_files = out_folder.get_folder_list()
        self.logger.info(u"file list {}".format(list_of_files))

        # has output xml file, otherwise error
        if self._calc._OUTXML_FILE_NAME not in list_of_files:
            successful = False
            self.logger.error(
                u"XML out not found {}".format(self._calc._OUTXML_FILE_NAME))
        else:
            has_xml_outfile = True

        # check if all files expected are there for the calculation
        for filel in self._should_retrieve:
            if filel not in list_of_files:
                successful = False
                self.logger.warning(
                    u"{} file not found in retrived folder, it"
                    " was probable not created by fleur".format(filel))

        # check if something was written to the error file
        if self._calc._ERROR_FILE_NAME in list_of_files:
            errorfile = os.path.join(out_folder.get_abs_path('.'),
                                     self._calc._ERROR_FILE_NAME)
            # read
            error_file_lines = ''
            try:
                with open(errorfile, 'r') as efile:# encoding="utf-8"
                    error_file_lines = efile.read()  # Note: read(), not readlines()
            except IOError:
                self.logger.error(
                    u"Failed to open error file: {}.".format(errorfile))

            # if not empty, has_error equals True, parse error.
            if error_file_lines:
                self.logger.warning(
                    u"The following was written into std error and piped to {} : \n {}"
                    "".format(self._calc._ERROR_FILE_NAME, unicode(error_file_lines.replace("'", ""), encoding="utf-8")))

                if 'OK' in error_file_lines: # if judft-error # TODO maybe change.
                    successful = True
                else:
                    successful = False

        if successful == False:
            return successful, ()

        #what about other files?
        #check input dict

        if self._calc._DOS_FILE_NAME in list_of_files:
            has_dos = True
        if self._calc._BAND_FILE_NAME in list_of_files:
            has_bands = True

        #if a new inp.xml file was created (new stucture)
        if self._calc._NEW_XMlINP_FILE_NAME in list_of_files:
            self.logger.error(u"new inp.xml file found in retrieved folder")
            has_new_xmlinp_file = True
        # add files which should also be there in addition to default_files.

        ################################
        ####### Parse the files ########

        if has_xml_outfile:
            # get outfile path and call xml out parser
            outxmlfile = os.path.join(
                out_folder.get_abs_path('.'), self._calc._OUTXML_FILE_NAME)
            simpledata, complexdata, parser_info, success = parse_xmlout_file(outxmlfile)
            if not success:
                # if the parsing is not smoothly we mark the calculation as FAILED...
                successful = False
                
            # Call routines for output node creation
            if simpledata:
                outputdata = dict(simpledata.items() + parser_info.items())
                outxml_params = ParameterData(dict=outputdata)
                link_name = self.get_linkname_outparams()# accessible via c.res
                new_nodes_list.append((link_name, outxml_params))

            if complexdata:
                parameter_data = dict(complexdata.items() + parser_info.items())
                outxml_params_complex = ParameterData(dict=parameter_data)
                link_name = self.get_linkname_outparams_complex()
                new_nodes_list.append((link_name, outxml_params_complex))

            #greate new fleurinpData object if needed

            #fleurinp_Data = FleurinpData(files= [inpxmlfile])
            #, symoutfile, enparafile])
            #self.logger.info('FleurinpData initialized')
            #self.logger.info
            #link_name_fleurinp = 'fleurinpData'

            # return it to the execmanager / maybe later not
            #new_nodes_list.append((link_name_fleurinp, fleurinp_Data))

            # load old fleurinpdata

            # if structure changed, create new fleurinpdata of new inp.xml file.
            # and parse the structure
 
        # optional parse other files

        # DOS
        if has_dos_file:
            dos_file = os.path.join(
                out_folder.get_abs_path('.'), self._calc._DOS_FILE_NAME)
            #if dos_file is not None:
            try:
                with open(dos_file, 'r') as dosf:#, encoding="utf-8"
                    dos_lines = dosf.read()  # Note: read() and not readlines()
            except IOError:
                raise FleurOutputParsingError(
                    "Failed to open DOS file: {}.".format(dos_file))
            dos_data = parse_dos_file(dos_lines)#, number_of_atom_types)

            # save array

        # Bands
        if has_bands_file:
            # TODO be carefull there might be two files.
            band_file = os.path.join(
                out_folder.get_abs_path('.'), self._calc._BAND_FILE_NAME)

            #if band_file is not None:
            try:
                with open(band_file, 'r') as bandf: #, encoding="utf-8"
                    bands_lines = bandf.read()  # Note: read() and not readlines()
            except IOError:
                raise FleurOutputParsingError(
                          "Failed to open bandstructure file: {}."
                          "".format(band_file))
            bands_data = parse_bands_file(bands_lines)

                # save array
        if has_new_xmlinp_file:
            new_inpxmlfile = os.path.join(
                out_folder.get_abs_path('.'), self._calc._NEW_XMlINP_FILE_NAME)
            new_fleurinpData = FleurinpData()
            new_fleurinpData.set_file(new_inpxmlfile, dst_filename= 'inp.xml')
            self.logger.info(u'New FleurinpData initialized')
            link_name = 'fleurinpData'#self.get_linkname_outparams()# accessible via c.res
            new_nodes_list.append((link_name, new_fleurinpData))

        # Spectra

        return successful, new_nodes_list

def parse_xmlout_file(outxmlfile):
    """
    Parses the out.xml file of a FLEUR calculation
    Receives as input the absolut path to the xml output file

    :param outxmlfile: path to out.xml file

    :returns xml_data_dict: a simple dictionary (QE output like)
                            with parsed data

    :raises FleurOutputParsingError: for errors in the parsing?
    """
    from lxml import etree#, objectify
    #from lxml.etree import XMLSyntaxError

    global parser_info_out

    parser_info_out = {'parser_warnings': [], 'unparsed' : []}
    parser_version = '0.1beta'
    parser_info_out['parser_info'] = 'AiiDA Fleur Parser v{}'.format(parser_version)
    #parsed_data = {}

    successful = True
    outfile_broken = False
    parse_xml = True
    parser = etree.XMLParser(recover=False)#, remove_blank_text=True)

    try:
        tree = etree.parse(outxmlfile, parser)
    except etree.XMLSyntaxError:
        outfile_broken = True
        parser_info_out['parser_warnings'].append(
            'The out.xml file is broken I try to repair it. Also I mark the parsing as failed.')
        successful = False

    if outfile_broken:
        #repair xmlfile and try to parse what is possible.
        parser = etree.XMLParser(recover=True)#, remove_blank_text=True)
        try:
            tree = etree.parse(outxmlfile, parser)
        except etree.XMLSyntaxError:
            #raise FleurOutputParsingError(
            #    "Failed to parse broken xml file: {}.".format(xml_file))
            parser_info_out['parser_warnings'].append(
                'Skipping the parsing of the xml file. '
                'Repairing was not possible.')
            parse_xml = False
            successful = False

    def parse_simplexmlout_file(root, outfile_broken):
        """
        Parses the xml.out file of a Fleur calculation
        Receives in input the root of an xmltree of the xml output file

        :param xmltree: root node of etree of out.xml file

        :returns xml_data_dict: a simple dictionary (QE output like)
                                with parsed data

        :raises FleurOutputParsingError: for errors in the parsing?
        """

        ### all xpath used. (maintain this) ###
        iteration_xpath = '/fleurOutput/scfLoop/iteration'
        magnetism_xpath = '/fleurOutput/inputData/calculationSetup/magnetism'

        relPos_xpath = '/fleurOutput/inputData/atomGroups/atomGroup/relPos'
        absPos_xpath = '/fleurOutput/inputData/atomGroups/atomGroup/absPos'
        filmPos_xpath = '/fleurOutput/inputData/atomGroups/atomGroup/filmPos'

        atomstypes_xpath = '/fleurOutput/inputData/atomGroups/atomGroup'
        symmetries_xpath = '/fleurOutput/inputData/cell/symmetryOperations/symOp'
        kpoints_xpath = '/fleurOutput/inputData/calculationSetup/bzIntegration/kPointList/kPoint'
        species_xpath = '/fleurOutput/inputData/atomSpecies'

        # input parameters
        creator_name_xpath = 'programVersion/@version'
        output_version_xpath = '/fleurOutput/@fleurOutputVersion'
        creator_target_architecture_xpath = 'programVersion/targetComputerArchitectures/text()'
        creator_target_structure_xpath = 'programVersion/targetStructureClass/text()'
        precision_xpath = 'programVersion/precision/@type'

        title_xpath = '/fleurOutput/inputData/comment/text()'
        kmax_xpath = 'calculationSetup/cutoffs'
        gmax_xpath = 'calculationSetup/cutoffs'
        mixing_xpath = 'calculationSetup/scfLoop'
        number_of_bands_xpath = 'calculationSetup/cutoffs'
        #lda_plus_u_calculation_xpath = ''
        #non_colinear_calculation_xpath = ''
        #inversion_symmetry = ''
        #coretail_correction = ''
        #xc_relatvistic_correction =''
        #lapw_basis_size =''
        spin_orbit_calculation = 'calculationSetup/soc'
        smearing_energy_xpath = 'calculationSetup/bzIntegration/@fermiSmearingEnergy'
        jspin_name = 'jspins'

        # timing
        start_time_xpath  = '/fleurOutput/startDateAndTime/@time'
        end_time_xpath = '/fleurOutput/endDateAndTime/@time'
        start_date_xpath  = '/fleurOutput/startDateAndTime/@date'
        end_date_xpath = '/fleurOutput/endDateAndTime/@date'



        ###########

        # get all iterations in out.xml file
        iteration_nodes = eval_xpath2(root, iteration_xpath)
        nIteration = len(iteration_nodes)
        data_exists = True

        # parse only last stable interation
        # (if modes (dos and co) maybe parse anyway if broken?)
        if outfile_broken and (nIteration >= 2):
            iteration_to_parse = iteration_nodes[-2]
            parser_info_out['last_iteration_parsed'] = nIteration-2
        elif outfile_broken and (nIteration == 1):
            iteration_to_parse = iteration_nodes[0]
            parser_info_out['last_iteration_parsed'] = nIteration
        elif not outfile_broken and (nIteration >= 1):
            iteration_to_parse = iteration_nodes[-1]
        else: # dont parse? # there was no iteration found.
            # maybe only the starting charge density was calculated
            parser_info_out['parser_warnings'].append(
                'There was no iteration found in the outfile, either just a '
                'starting density was generated or something went wrong.')
            data_exists = False
            iteration_to_parse = None

        # for getting the fleur modes use fleurinp methods
        spin = get_xml_attribute(eval_xpath(root, magnetism_xpath), jspin_name)
        if spin:
            Fleurmode = {'jspin' : int(spin)}
        else:
            Fleurmode = {'jspin' : 1}
        if data_exists:
            simple_data = parse_simple_outnode(iteration_to_parse, Fleurmode)
        else:
            simple_data = {}

        # TODO int he future add here the warnings retunred from parse_simple_outnode
        # Currently Fleur warnings an errors are not written to the out.xml
        # should they be lists or dicts
        warnings={'info': {}, 'debug' : {}, 'warning' : {}, 'error' : {}}

        simple_data['number_of_atoms'] = (len(eval_xpath2(root, relPos_xpath)) +
                                          len(eval_xpath2(root, absPos_xpath)) +
                                          len(eval_xpath2(root, filmPos_xpath)))
        simple_data['number_of_atom_types'] = len(eval_xpath2(root, atomstypes_xpath))
        simple_data['number_of_iterations'] = nIteration
        simple_data['number_of_symmetries'] = len(eval_xpath2(root, symmetries_xpath))
        simple_data['number_of_species'] = len(eval_xpath2(root, species_xpath))
        simple_data['number_of_kpoints'] = len(eval_xpath2(root, kpoints_xpath))
        simple_data['number_of_spin_components'] = Fleurmode['jspin']

        title = eval_xpath(root, title_xpath)
        if title:
            title = u'{}'.format(title).encode('utf-8').strip() #str(title).strip()
        simple_data['title'] = title
        simple_data['creator_name'] = eval_xpath(root, creator_name_xpath)
        simple_data['creator_target_architecture'] = eval_xpath(root, creator_target_architecture_xpath)
        simple_data['creator_target_structure'] = eval_xpath(root, creator_target_structure_xpath)
        simple_data['output_file_version'] = eval_xpath(root, output_version_xpath)




        # time
        # Maybe change the behavior if things could not be parsed...
        # Especilly if file was broken, ie endtime it not there.
        starttime = eval_xpath(root, start_time_xpath)
        if starttime:
            starttimes = starttime.split(':')
        else:
            starttimes = [0,0,0]
            msg = u'Startime was unparsed, inp.xml prob not complete, do not believe the walltime!'
            if data_exists:
                parser_info_out['parser_warnings'].append(msg)

        endtime = eval_xpath(root, end_time_xpath)
        if endtime:
            endtimes = endtime.split(':')
        else:
            endtimes = [0,0,0]
            msg = 'Endtime was unparsed, inp.xml prob not complete, do not believe the walltime!'
            if data_exists:
                parser_info_out['parser_warnings'].append(msg)
        start_date = eval_xpath(root, start_date_xpath)
        end_date = eval_xpath(root, end_date_xpath)

        offset = 0
        if start_date != end_date:
            # date="2018/01/15", Can this fail? what happens if not there
            if start_date and end_date:
                date_sl = [int(ent) for ent in start_date.split('/')]
                date_el = [int(ent) for ent in end_date.split('/')]
                date_s = date(*date_sl)
                date_e = date(*date_el)
                diff = date_e - date_s
                offset = diff.days * 86400
        #ncores = 12 #TODO parse parallelization_Parameters
        time = offset + (int(endtimes[0])-int(starttimes[0]))*60*60 + (int(endtimes[1])-int(starttimes[1]))*60 + int(endtimes[2]) - int(starttimes[2])
        simple_data['walltime'] = time
        simple_data['walltime_units'] = 'seconds'
        #simple_data['core_hours'] = time*ncores*1.0/3600
        #simple_data['parallelization_Parameters'] = {'mpiPEs' : ncores}
        simple_data['start_date'] = {'date' : start_date, 'time' : starttime}
        simple_data['end_date'] = {'date' : end_date, 'time' : endtime}

        warnings['info'] = {}#TODO
        warnings['debug'] = {} #TODO
        warnings['warning'] = {}#TODO
        warnings['error'] = {}#TODO
        simple_data['warnings'] = warnings

        return simple_data

    # TODO find a way to import these from xml_util, but make the parser logger work...
    def eval_xpath(node, xpath):
        """
        Tries to evalutate an xpath expression. If it fails it logs it.

        :param root node of an etree and an xpath expression (relative, or absolute)
        :returns either nodes, or attributes, or text
        """
        try:
            return_value = node.xpath(xpath)
        except etree.XPathEvalError:
            parser_info_out['parser_warnings'].append(
                'There was a XpathEvalError on the xpath: {} \n Either it does '
                'not exist, or something is wrong with the expression.'
                ''.format(xpath))
            return []# or rather None?
        if len(return_value) == 1:
            return return_value[0]
        else:
            return return_value

    def eval_xpath2(node, xpath):
        """
        Tries to evalutate an xpath expression. If it fails it logs it.

        :param root node of an etree and an xpath expression (relative, or absolute)
        :returns ALWAYS a node list
        """
        try:
            return_value = node.xpath(xpath)
        except etree.XPathEvalError:
            parser_info_out['parser_warnings'].append(
                'There was a XpathEvalError on the xpath: {} \n Either it does '
                'not exist, or something is wrong with the expression.'
                ''.format(xpath))
            return []
        if len(return_value) == 1:
            return return_value
        else:
            return return_value

    def get_xml_attribute(node, attributename):
        """
        Get an attribute value from a node.

        :param node: a node from etree
        :param attributename: a string with the attribute name.
        :returns either attributevalue, or None
        """
        if etree.iselement(node):
            attrib_value = node.get(attributename)
            if attrib_value:
                return attrib_value
            else:
                parser_info_out['parser_warnings'].append(
                    'Tried to get attribute: "{}" from element {}.\n '
                    'I recieved "{}", maybe the attribute does not exist'
                    ''.format(attributename, node, attrib_value))
                return None
        else: # something doesn't work here, some nodes get through here
            parser_info_out['parser_warnings'].append(
                'Can not get attributename: "{}" from node "{}", '
                'because node is not an element of etree.'
                ''.format(attributename, node))

            return None


    def convert_to_float(value_string):
        """
        Tries to make a float out of a string. If it can't it logs a warning
        and returns True or False if convertion worked or not.

        :param value_string: a string
        :returns value: the new float or value_string: the string given
        :retruns True or Falses
        """
        # TODO lowercase everything
        try:
            value = float(value_string)
        except TypeError:
            parser_info_out['parser_warnings'].append(
                'Could not convert: "{}" to float, TypeError'
                ''.format(value_string))
            return value_string, False
        except ValueError:
            parser_info_out['parser_warnings'].append(
                'Could not convert: "{}" to float, ValueError'
                ''.format(value_string))
            return value_string, False
        return value, True

    def convert_to_int(value_string):
        """
        Tries to make a int out of a string. If it can't it logs a warning
        and returns True or False if convertion worked or not.

        :param value_string: a string
        :returns value: the new int or value_string: the string given
        :retruns True or False
        """
        try:
            value = int(value_string)
        except TypeError:
            parser_info_out['parser_warnings'].append(
                'Could not convert: "{}" to int, TypeError'
                ''.format(value_string))
            return value_string, False
        except ValueError:
            parser_info_out['parser_warnings'].append(
                'Could not convert: "{}" to int, ValueError'
                ''.format(value_string))
            return value_string, False
        return value, True


    def convert_htr_to_ev(value):
        """
        Multiplies the value given with the Hartree factor (converts htr to eV)
        """
        htr = 27.21138602
        suc = False
        value_to_save, suc = convert_to_float(value)
        if suc:
            return value_to_save*htr
        else:
            return value

    def parse_simple_outnode(iteration_node, Fleurmode):
        """
        Parses the data from the iteration given (usually last iteration)
        and some other data for the 'simple' output node.

        :param iteration_node: etree node of an interation
        :param Fleurmode: python dictionary, with all the modes,
                          which influence the parsing

        :returns simple_data: a python dictionary with all results
        """
        ###################################################
        ##########  all xpaths (maintain this) ############
        # (specifies where to find things in the out.xml) #

        # density
        densityconvergence_xpath = 'densityConvergence'
        chargedensity_xpath = 'densityConvergence/chargeDensity'
        overallchargedensity_xpath = 'densityConvergence/overallChargeDensity'
        spindensity_xpath = 'densityConvergence/spinDensity'

        bandgap_xpath = 'bandgap'
        fermi_energy_xpath = 'FermiEnergy'

        #magnetic moments
        magnetic_moments_in_mtpheres_xpath = 'magneticMomentsInMTSpheres'
        magneticmoment_xpath = 'magneticMomentsInMTSpheres/magneticMoment'

        magneticmoments_xpath = 'magneticMomentsInMTSpheres/magneticMoment/@moment'
        magneticmoments_spinupcharge_xpath = 'magneticMomentsInMTSpheres/magneticMoment/@spinUpCharge'
        magneticmoments_spindowncharge_xpath = 'magneticMomentsInMTSpheres/magneticMoment/@spinDownCharge'

        orbmagnetic_moments_in_mtpheres_xpath = 'orbitalMagneticMomentsInMTSpheres'
        orbmagneticmoment_xpath = 'orbitalMagneticMomentsInMTSpheres/orbMagMoment'

        orbmagneticmoments_xpath = 'orbitalMagneticMomentsInMTSpheres/orbMagMoment/@moment'
        orbmagneticmoments_spinupcharge_xpath = 'orbitalMagneticMomentsInMTSpheres/orbMagMoment/@spinUpCharge'
        orbmagneticmoments_spindowncharge_xpath = 'orbitalMagneticMomentsInMTSpheres/orbMagMoment/@spinDownCharge'


        spinupcharge_name = 'spinUpCharge'
        spindowncharge_name = 'spinDownCharge'
        moment_name = 'moment'


        # all electron charges

        allelectronchages_xpath = ''

        a = 'total'
        b = 'interstitial'
        c ='value'
        # energy
        totalenergy_xpath = 'totalEnergy'
        sumofeigenvalues_xpath = 'totalEnergy/sumOfEigenvalues'
        core_electrons_xpath = 'totalEnergy/sumOfEigenvalues/coreElectrons'
        valence_electrons_xpath = 'totalEnergy/sumOfEigenvalues/valenceElectrons'
        chargeden_xc_den_integral_xpath = 'totalEnergy/chargeDenXCDenIntegral'

        #free_energy_xpath = 'totalEnergy/freeEnergy'

        # forces
        forces_units_xpath = 'totalForcesOnRepresentativeAtoms'
        forces_total_xpath = 'totalForcesOnRepresentativeAtoms/forceTotal'

        #
        iteration_xpath = '.'


        units_name = 'units'
        value_name = 'value'
        distance_name = 'distance'

        overall_number_name = 'overallNumber'
        atomtype_name = 'atomType'

        # forces

        f_x_name = 'F_x'
        f_y_name = 'F_y'
        f_z_name = 'F_z'
        new_x_name = 'x'
        new_y_name = 'y'
        new_z_name = 'z'
        ###################################################

        jspin = Fleurmode['jspin']
        simple_data = {}


        def write_simple_outnode(value, value_type, value_name, dict):
            """
            writes a value (int or float) in the simple data dict.
            if path does not exit it initializes it!

            if the value has not the type given,
            it will be logged in parser info instead of writing.

            :param value: value
            """

            interation_current_number_name = 'numberForCurrentRun'
            suc = False

            if value_type == 'float':
                value_to_save, suc = convert_to_float(value)
            elif value_type == 'int':
                value_to_save, suc = convert_to_int(value)
            elif value_type == 'str':
                suc = True
                value_to_save = u'{}'.format(value).encode('utf-8') #.strip() #str(title).strip()
                #value_to_save, suc = convert_to_str(value)
            elif value_type =='list':
                suc = True
                value_to_save = value
            elif value_type =='list_floats':
                value_to_save = []
                for val in value:
                    value_to_savet, suct = convert_to_float(val)
                    value_to_save.append(value_to_savet)
                suc = True # TODO individual or common error message?
            elif value_type =='list_list_floats':
                value_to_save = []
                for val in value:
                    value_to_savet = []
                    for val1 in val:
                        value_to_savet1, suct = convert_to_float(val1)
                        value_to_savet.append(value_to_savet1)
                    value_to_save.append(value_to_savet)
                suc = True # TODO individual or common error message?
            else:
                #self.logger.error('I dont know the type you gave me {}'.format(type))
                pass
                # TODO log error, self is not known here...
            if suc:
                dict[value_name] = value_to_save
            else:
                parser_info_out['unparsed'].append(
                    {value_name : value,
                     'iteration' : get_xml_attribute(iteration_node, interation_current_number_name)})


        # total energy
        units_e = get_xml_attribute(
            eval_xpath(iteration_node, totalenergy_xpath), units_name)
        write_simple_outnode(
            units_e, 'str', 'energy_hartree_units', simple_data)

        tE_htr = get_xml_attribute(
            eval_xpath(iteration_node, totalenergy_xpath), value_name)
        write_simple_outnode(tE_htr, 'float', 'energy_hartree', simple_data)

        write_simple_outnode(
            convert_htr_to_ev(tE_htr), 'float', 'energy', simple_data)
        write_simple_outnode('eV', 'str', 'energy_units', simple_data)

        sumofeigenvalues = get_xml_attribute(
            eval_xpath(iteration_node, sumofeigenvalues_xpath), value_name)
        write_simple_outnode(
            sumofeigenvalues, 'float', 'sum_of_eigenvalues', simple_data)

        coreElectrons = get_xml_attribute(
            eval_xpath(iteration_node, core_electrons_xpath), value_name)
        write_simple_outnode(
            coreElectrons, 'float', 'energy_core_electrons', simple_data)

        valenceElectrons = get_xml_attribute(
            eval_xpath(iteration_node, valence_electrons_xpath), value_name)
        write_simple_outnode(
            valenceElectrons, 'float', 'energy_valence_electrons', simple_data)

        ch_d_xc_d_inte = get_xml_attribute(
            eval_xpath(iteration_node, chargeden_xc_den_integral_xpath), value_name)
        write_simple_outnode(
            ch_d_xc_d_inte, 'float', 'charge_den_xc_den_integral', simple_data)

        # bandgap
        units_bandgap = get_xml_attribute(
            eval_xpath(iteration_node, bandgap_xpath), units_name)
        write_simple_outnode(units_bandgap, 'str', 'bandgap_units', simple_data)

        bandgap = get_xml_attribute(
            eval_xpath(iteration_node, bandgap_xpath), value_name)
        write_simple_outnode(bandgap, 'float', 'bandgap', simple_data)

        # fermi
        fermi_energy = get_xml_attribute(
            eval_xpath(iteration_node, fermi_energy_xpath), value_name)
        write_simple_outnode(fermi_energy, 'float', 'fermi_energy', simple_data)
        units_fermi_energy = get_xml_attribute(
            eval_xpath(iteration_node, fermi_energy_xpath), units_name)
        write_simple_outnode(
            units_fermi_energy, 'str', 'fermi_energy_units', simple_data)

        # density convergence
        units = get_xml_attribute(
            eval_xpath(iteration_node, densityconvergence_xpath), units_name)
        write_simple_outnode(
            units, 'str', 'density_convergence_units', simple_data)


        if jspin == 1:
            charge_density = get_xml_attribute(
                eval_xpath(iteration_node, chargedensity_xpath), distance_name)
            write_simple_outnode(
                charge_density, 'float', 'charge_density', simple_data)

        elif jspin == 2:
            charge_densitys = eval_xpath(iteration_node, chargedensity_xpath)

            if charge_densitys:# otherwise we get a keyerror if calculation failed.
                charge_density1 = get_xml_attribute(charge_densitys[0], distance_name)
                charge_density2 = get_xml_attribute(charge_densitys[1], distance_name)
            else: # Is non a problem?
                charge_density1 = None
                charge_density2 = None
            write_simple_outnode(
                charge_density1, 'float', 'charge_density1', simple_data)
            write_simple_outnode(
                charge_density2, 'float', 'charge_density2', simple_data)

            spin_density = get_xml_attribute(
                eval_xpath(iteration_node, spindensity_xpath), distance_name)
            write_simple_outnode(
                spin_density, 'float', 'spin_density', simple_data)

            overall_charge_density = get_xml_attribute(
                eval_xpath(iteration_node, overallchargedensity_xpath), distance_name)
            write_simple_outnode(
                overall_charge_density, 'float', 'overall_charge_density', simple_data)

            # magnetic moments            #TODO orbMag Moment
            m_units = get_xml_attribute(
                eval_xpath(iteration_node, magnetic_moments_in_mtpheres_xpath), units_name)
            write_simple_outnode(
                m_units, 'str', 'magnetic_moment_units', simple_data)
            write_simple_outnode(
                m_units, 'str', 'orbital_magnetic_moment_units', simple_data)

            moments = eval_xpath(iteration_node, magneticmoments_xpath)
            write_simple_outnode(
                moments, 'list_floats', 'magnetic_moments', simple_data)

            spinup = eval_xpath(iteration_node, magneticmoments_spinupcharge_xpath)
            write_simple_outnode(
                spinup, 'list_floats', 'magnetic_spin_up_charges', simple_data)

            spindown = eval_xpath(iteration_node, magneticmoments_spindowncharge_xpath)
            write_simple_outnode(
                spindown, 'list_floats', 'magnetic_spin_down_charges', simple_data)

            #orbital magnetic moments
            orbmoments = eval_xpath(iteration_node, orbmagneticmoments_xpath)
            write_simple_outnode(
                orbmoments, 'list_floats', 'orbital_magnetic_moments', simple_data)

            orbspinup = eval_xpath(iteration_node, orbmagneticmoments_spinupcharge_xpath)
            write_simple_outnode(
                orbspinup, 'list_floats', 'orbital_magnetic_spin_up_charges', simple_data)

            orbspindown = eval_xpath(iteration_node, orbmagneticmoments_spindowncharge_xpath)
            write_simple_outnode(
                orbspindown, 'list_floats', 'orbital_magnetic_spin_down_charges', simple_data)

            # TODO atomtype dependence
            #moment = get_xml_attribute(
            #    eval_xpath(iteration_node, magneticmoment_xpath), moment_name)
            #write_simple_outnode(moment, 'float', 'magnetic_moment', simple_data)

            #spinup = get_xml_attribute(
            #    eval_xpath(iteration_node, magneticmoment_xpath), spinupcharge_name)
            #write_simple_outnode(spinup, 'float', 'spin_up_charge', simple_data)

            #spindown = get_xml_attribute(
            #    eval_xpath(iteration_node, magneticmoment_xpath), spindowncharge_name)
            #write_simple_outnode(spindown, 'float', 'spin_down_charge', simple_data)

            #Total charges, total magentic moment

        # total iterations
        number_of_iterations_total = get_xml_attribute(
            eval_xpath(iteration_node, iteration_xpath), overall_number_name)
        write_simple_outnode(
            number_of_iterations_total, 'int', 'number_of_iterations_total', simple_data)


        #forces atomtype dependend
        forces = eval_xpath2(iteration_node, forces_total_xpath)
        # length should be ntypes
        largest_force = -0.0
        for force in forces:
            atomtype, success = convert_to_int(get_xml_attribute(force, atomtype_name))

            forces_unit = get_xml_attribute(
                eval_xpath(iteration_node, forces_units_xpath), units_name)
            write_simple_outnode(forces_unit, 'str', 'force_units', simple_data)

            force_x = get_xml_attribute(force, f_x_name)
            write_simple_outnode(
                force_x, 'float', 'force_x_type{}'.format(atomtype), simple_data)

            force_y = get_xml_attribute(force, f_y_name)
            write_simple_outnode(
                force_y, 'float', 'force_y_type{}'.format(atomtype), simple_data)

            force_z = get_xml_attribute(force, f_z_name)
            write_simple_outnode(
                force_z, 'float', 'force_z_type{}'.format(atomtype), simple_data)


            force_xf, suc1 = convert_to_float(force_x)
            force_yf, suc2 = convert_to_float(force_y)
            force_zf, suc3 = convert_to_float(force_z)

            if suc1 and suc2 and suc3:
                if abs(force_xf) > largest_force:
                    largest_force = abs(force_xf)
                if abs(force_yf) > largest_force:
                    largest_force = abs(force_yf)
                if abs(force_zf) > largest_force:
                    largest_force = abs(force_zf)

            pos_x = get_xml_attribute(force, new_x_name)
            write_simple_outnode(
                pos_x, 'float', 'abspos_x_type{}'.format(atomtype), simple_data)
            pos_y = get_xml_attribute(force, new_y_name)
            write_simple_outnode(
                pos_y, 'float', 'abspos_y_type{}'.format(atomtype), simple_data)
            pos_z = get_xml_attribute(force, new_z_name)
            write_simple_outnode(
                pos_z, 'float', 'abspos_z_type{}'.format(atomtype), simple_data)

        write_simple_outnode(
            largest_force, 'float', 'force_largest', simple_data)

        return simple_data


    if parse_xml:
        root = tree.getroot()
        simple_out = parse_simplexmlout_file(root, outfile_broken)
        simple_out['outputfile_path'] = outxmlfile
        #TODO parse complex out
        complex_out = {} #parse_xmlout_file(root)
        return simple_out, complex_out, parser_info_out, successful
    else:
        return {}, {}, parser_info_out, successful



def parse_dos_file(dos_lines):#, number_of_atom_types):
    """
    Parses the returned DOS.X files.
    Structure:
    (100(1x,e10.3))   e,totdos,interstitial,vac1,vac2,
    (at(i),i=1,ntype),((q(l,i),l=1,LMAX),i=1,ntype)
    where e is the energy in eV (= 1/27.2 htr) at(i) is the local DOS of a
    single atom of the i'th atom-type and q(l,i) is the l-resolved DOS at
    the i'th atom but has to be multiplied by the number of atoms of this type.

    :param dos_lines: string of the read in dos file
    :param number_of_atom_types: integer, number of atom types
    """
    #pass
    return 0

def parse_bands_file(bands_lines):
    '''
    Parses the returned bands.1 and bands.2 file and returns a complete
    bandsData object. bands.1 has the form: k value, energy

    :param bands_lines: string of the read in bands file
    '''
    # TODO not finished
    # read bands out of file:
    nrows = 0 # get number of rows (known form number of atom types
    bands_values = [] #init an array of arrays nkpoint * ...
    bands_labels = [] # label for each row.

    # fill and correct fermi energy.
    bands_values = []


    # TODO we need to get the cell from StructureData node
    # and KpointsData node from inpxml
    fleur_bands = BandsData()
    #fleur_bands.set_cell(cell)
    #fleur_bands.set_kpoints(kpoints, cartesian=True)
    fleur_bands.set_bands(bands=bands_values,
                          units='eV',
                          labels=bands_labels)

    for line in bands_lines:
        pass
    return fleur_bands




# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################
"""
In this module is the FleurinpData class, and methods for FLEUR input
manipulation plus methods for extration of AiiDA data structures.
"""
# TODO: this needs to be cleaned up and redocumented
# TODO: all methods to change now in fleurinpmodifier, do we still want to
# store the userchanges, or rather delete them? depends how well one can see things
# from fleurinpmodifier...
# TODO: maybe add a modify method which returns a fleurinpmodifier class
# TODO: inpxml to dict: maybe kpts should not be writen to the dict? same with symmetry
# TODO: test for large input files, I believe the recursion is still quite slow..
# TODO: 2D cell get kpoints and get structure also be carefull with tria = T!!!
#TODO : maybe save when get_structure or get_kpoints was executed on fleurinp,
# because otherwise return this node instead of creating a new one!

import os
import re
from lxml import etree
#from lxml.etree import XMLSyntaxError

from aiida.orm import Data
from aiida.common.exceptions import InputValidationError, ValidationError
from aiida_fleur.tools.xml_util import xml_set_attribv_occ, xml_set_first_attribv
from aiida_fleur.tools.xml_util import  xml_set_all_attribv, xml_set_text, replace_tag
from aiida.work import workfunction as wf
from aiida_fleur.fleur_schema.schemafile_index import get_internal_search_paths, get_schema_paths

bohr_a = 0.52917721092#A, todo: get from somewhereA


class FleurinpData(Data):
    """
    AiiDA data object representing everything a FLEUR calculation needs.

    It is initialized with an absolute path to an inp.xml file.
    Other files can also be added, which will be compied to the remote machine, where the calculation takes place.

    It stores the files in the repository and stores the input parameters of the
    inp.xml file of FLEUR in the Database as a python dictionary (as internal attributes).
    When an inp.xml (name important!) file is added to files, FleurinpData searches
    for a corresponding xml schema file in the PYTHONPATH environment variable.
    Therefore, it is recommened to have the plug-in source code directory in the python environment..
    If no corresponding schema file is found an error is raised.

    FleurinpData further provides the user with
    methods to extract AiiDA StructureData and KpointsData nodes.

    Remember that most attributes of AiiDA nodes can not be changed after they
    have been stored in the DB! Therefore, you have to use the FleurinpModifier class and its methods
    if you want to change somthing in the inp.xml file. You will retrieve a new FLeurinpdata that way and
    start a new calculation from it.
    """

    # serach in current folder and search in aiida source code
    # we want to search in the Aiida source directory, get it from python path,
    # maybe better from somewhere else.
    #TODO: dont walk the whole python path, test if dir below is aiida?
    #needs to be imporved, schema file is often after new installation not found...
    #installation with pip should always lead to a schmea file in the python path, or even specific place

    _search_paths = []
    ifolders = get_internal_search_paths()
    ischemas = get_schema_paths()
    for path in ischemas:
        _search_paths.append(path)
    for path in ifolders:
        _search_paths.append(path)
    _search_paths.append('./')

    # Now add also python path maybe will be decaptivated
    #if pythonpath is non existant catch error
    try:
        pythonpath = os.environ['PYTHONPATH'].split(':')
    except KeyError:
        pythonpath = []

    for path in pythonpath[:]:
        _search_paths.append(path)

    #_search_paths = ['./', '/Users/broeder/aiida/codes/fleur/',
    #                 str(get_repository_folder())]

    @property
    def _has_schema(self):
        """
        Boolean property, which stores if a schema file is already known
        """
        return self.get_attr('_has_schema', False)

    @property
    def _schema_file_path(self):
        """
        A string, which stores the absolute path to the schemafile fount
        """
        return self.get_attr('_schema_file_path', None)

    @_has_schema.setter
    def _has_schema(self, boolean):
        """
        Setter for has_schema
        """
        self._set_attr('_has_schema', boolean)

    @_schema_file_path.setter
    def _schema_file_path(self, schemapath):
        """
        Setter for the schema file path
        """
        self._set_attr('_schema_file_path', schemapath)

    # files
    @property
    def files(self):
        """
        Returns the list of the names of the files stored
        """
        return self.get_attr('files', [])

    @files.setter
    def files(self, filelist):
        """
        Add a list of files to FleurinpData.
        Alternative use setter method.

        :param files: list of filepaths
        """
        for file1 in filelist:
            self.set_file(file1)

    def set_files(self, files):
        """
        Add a list of files to FleurinpData
        Alternative setter

        :param files: list of filepaths
        """
        self.files = files

    def set_file(self, filename, dst_filename=None):
        """
        Add a file to the FleurinpData

        :param filename: absolute path to the file
        """
        self._add_path(filename, dst_filename=dst_filename)

    def del_file(self, filename):
        """
        Remove a file from FleurinpData

        :param filename: name of the file stored in the DB
        """
        # remove from files attr list
        if filename in self.get_attr('files'):
            try:
                self.get_attr('files').remove(filename)
                #self._del_attr('filename')
            except AttributeError:
                ## There was no file set
                pass
        # remove from sandbox folder
        if filename in self.get_folder_list():
            super(FleurinpData, self).remove_path(filename)

    def get_file_abs_path(self, filename):
        """
        Return the absolute path to a file in the repository

        :param filename: name of the file

        """
        if filename in self.files:
            return os.path.join(self._get_folder_pathsubfolder.abspath, filename)
        else:
            raise ValueError(
                '{} is not in {}'.format(filename,
                    os.path.join(self._get_folder_pathsubfolder.abspath)))

    def find_schema(self, inp_version_number):
        """
        Method which searches for a schema files (.xsd) which corresponds
        to the input xml file. (compares the version numbers)
        """
        # user changed version number, or no file yet known.
        # TODO test if this still does the right thing if user adds one
        #inp.xml and then an inp.xml with different version.
        # or after copying and read
        schemafile_paths = []

        for path in self._search_paths:#paths:
            for root, dirs, files in os.walk(path):
                for file1 in files:
                    if file1.endswith(".xsd"):
                        if ('Fleur' in file1) or ('fleur' in file1):
                            schemafile_path = os.path.join(root, file1)
                            schemafile_paths.append(schemafile_path)
                            i = 0
                            imin = 0
                            imax = 0
                            schemafile = open(schemafile_path, 'r')
                            for line in schemafile.readlines():
                                i = i + 1
                                # This kind of hardcoded
                                if re.search('name="FleurVersionType"', line):
                                    imax = i + 10 # maybe make larger or different
                                    imin = i
                                    schema_version_numbers = []
                                if (i > imin) and (i <= imax):
                                    if re.search('enumeration value', line):
                                        schema_version_number = re.findall(r'\d+.\d+', line)[0]
                                        schema_version_numbers.append(schema_version_number)
                                    elif re.search('simpleType>', line):
                                        break
                            schemafile.close()
                            #test if schemafiles works with multiple fleur versions
                            for version_number in schema_version_numbers:
                                if version_number == inp_version_number:
                                    #we found the right schemafile for the current inp.xml
                                    self._set_attr('_schema_file_path', schemafile_path)
                                    self._set_attr('_has_schema', True)
                                    return schemafile_paths, True

        return schemafile_paths, False

    def _add_path(self, src_abs, dst_filename=None):
        """
        Add a single file to folder

        """
        #TODO, only certain files should be allowed to be added
        #_list_of_allowed_files = ['inp.xml', 'enpara', 'cdn1', 'sym.out', 'kpts']

        old_file_list = self.get_folder_list()

        if not os.path.isabs(src_abs):
            raise ValueError("Pass an absolute path for src_abs: {}".format(src_abs))

        if not os.path.isfile(src_abs):
            raise ValueError("src_abs must exist and must be a single file: {}".format(src_abs))

        if dst_filename is None:
            final_filename = os.path.split(src_abs)[1]
        else:
            final_filename = dst_filename

        super(FleurinpData, self).add_path(src_abs, final_filename)

        old_files_list = self.get_attr('files', [])

        if final_filename not in old_file_list:
            old_files_list.append(final_filename)
        self._set_attr('files', old_files_list)

        # here this is hardcoded, might want to change? get filename from elsewhere
        if final_filename == 'inp.xml':
            #get input file version number
            inp_version_number = None
            inpfile = open(src_abs, 'r')
            for line in inpfile.readlines():
                if re.search('fleurInputVersion', line):
                    inp_version_number = re.findall(r'\d+.\d+', line)[0]
                    break
            inpfile.close()
            if inp_version_number == None: # we raise after file closure
                raise InputValidationError("No fleurInputVersion number found "
                    "in given input file: {}. "
                    "Please check if this is a valid fleur input file. "
                    "It can not be validated and I can not use it. ".format(src_abs))
            # search for Schema file with same version number
            schemafile_paths, found = self.find_schema(inp_version_number)


            if (self._schema_file_path is None) or (self.inp_userchanges.has_key('fleurInputVersion')): #(not)
                schemafile_paths, found = self.find_schema(inp_version_number)
                if not(found):
                    raise InputValidationError("No XML schema file (.xsd) with matching version number {} "
                        "to the inp.xml file was found. I have looked here: {} "
                        "and have found only these schema files for Fleur: {}. "
                        "I need this file to validate your input and to know the structure "
                        "of the current inp.xml file, sorry.".format(inp_version_number,
                                                          self._search_paths, schemafile_paths))
            if (not self._has_schema) and (self._schema_file_path is None):
                raise InputValidationError("No XML schema file (.xsd) with matching version number {} "
                    "to the inp.xml file was found. I have looked here: {} "
                    "and have found only these schema files for Fleur: {}. "
                    "I need this file to validate your input and to know the structure "
                    "of the current inp.xml file, sorry.".format(inp_version_number,
                                                          self._search_paths, schemafile_paths))

            # set inp dict of Fleurinpdata
            self._set_inp_dict()

    def _set_inp_dict(self):
        """
        Sets the inputxml_dict from the inp.xml file attached to FleurinpData

        1. get inp.xml strucutre/layout
        2. load inp.xml file
        3. call inpxml_to_dict
        4. set inputxml_dict
        """
        from aiida_fleur.tools.xml_util import get_inpxml_file_structure, inpxml_todict
        # get inpxml structure
        inpxmlstructure = get_inpxml_file_structure()

        # read xmlinp file into an etree with autocomplition from schema
        inpxmlfile = self.get_file_abs_path('inp.xml')

        xmlschema_doc = etree.parse(self._schema_file_path)
        xmlschema = etree.XMLSchema(xmlschema_doc)
        parser = etree.XMLParser(schema=xmlschema, attribute_defaults=True)
        #dtd_validation=True

        tree = etree.parse(inpxmlfile)#, parser)
        # there is a bug when validating at parsetime, therefore we only
        #validate at parse time if file is invalid, to get nice error message
        if not xmlschema.validate(tree):
            tree = etree.parse(inpxmlfile, parser)

        root = tree.getroot()

        # convert etree into python dictionary
        inpxml_dict = inpxml_todict(root, inpxmlstructure)

        # set inpxml_dict attribute
        self._set_attr('inp_dict', inpxml_dict)


    # tracing user changes
    @property
    def inp_userchanges(self):
        """
        Return the changes done by the user on the inp.xml file.
        """
        return self.get_attr('inp_userchanges', {})

    # dict with inp paramters parsed from inp.xml
    @property
    def inp_dict(self):
        """
        Returns the inp_dict (the representation of the inp.xml file) as it will
        or is stored in the database.
        """
        return self.get_attr('inp_dict', {})

    def copy(self):
        """
        Method to copy a FleurinpData and keep the proverance.
        (because currently that is not default)
        """
        # TODO: do to clone() behavior of aiida 1.0 might be obsolete
        # copy only the files not the user changes and so on.

        filepath = []
        for file in self.files:
            filepath.append(self.get_file_abs_path(file))

        new = FleurinpData(files=filepath)

        return new

    # TODO better validation? other files, if has a schema
    def _validate(self):
        """
        validation method. Check here for properties that have to be there, for
        a valid fleurinpData object.
        """
        #from aiida.common.exceptions import ValidationError

        super(FleurinpData, self)._validate()

        if 'inp.xml' in self.files:
            #has_inpxml = True # does nothing so far
            pass
        else:
            raise ValidationError('inp.xml file not in attribute "files". '
                                  'FleurinpData needs to have and inp.xml file!')


    '''
        try:
            has_inpxml = 'inp.xml' in self.files
        except AttributeError:
            raise ValidationError("attribute 'filename' not set.")

        if self.files != self.get_folder_list():
            raise ValidationError("The list of files in the folder does not "
                                  "match the 'files' attribute. "
                                  "_files='{}', content: {}".format(
                self.files, self.get_folder_list()))
    '''

    def get_fleur_modes(self):
        '''
        Retrieve information from the inp.xml file. 'Modes' are paths a FLEUR
        calculation can take, resulting in different output files, dependend on the input.
        i.e other files need to be copied before and after the calculation.
        common modes are: scf, jspin 2, dos, band, pot8, lda+U, eels, ...
        '''
        # TODO these should be retrieved by looking at the inpfile struture and
        # then setting the paths.
        # use methods from fleur parser...
        # For now they are hardcoded.
        #    'dos': '/fleurInput/output',
        #    'band': '/fleurInput/output',
        #    'pot8': '/fleurInput/calculationSetup/expertModes',
        #    'jspins': '/fleurInput/calculationSetup/magnetism',
        fleur_modes = {'jspins' : '', 'dos' : '', 'band' : '', 'pot8' : '', 'ldau' : '', 'forces' : ''}
        if 'inp.xml' in self.files:
            fleur_modes['jspins'] = self.inp_dict['calculationSetup']['magnetism']['jspins'] #['fleurInput']
            fleur_modes['dos'] = self.inp_dict['output']['dos']#'fleurInput']
            fleur_modes['band'] = self.inp_dict['output']['band']
            #fleur_modes['pot8'] = self.inp_dict['calculationSetup']['expertModes']['pot8'] # pot8 was removed from inp.xml
            fleur_modes['forces'] = self.inp_dict['calculationSetup']['geometryOptimization']['l_f']
            ldau = False # TODO test if ldau in inp_dict....
            fleur_modes['ldau'] = False
        return fleur_modes

    #@staticmethod
    def get_structuredata_nwf(fleurinp):
        """
        This routine return an AiiDA Structure Data type produced from the inp.xml
        file. not a workfunction

        :return: StructureData node, or None
        """
        from aiida.orm.data.structure import StructureData
        from aiida_fleur.tools.StructureData_util import rel_to_abs, rel_to_abs_f

        #Disclaimer: this routine needs some xpath expressions. these are hardcoded here,
        #therefore maintainance might be needed, if you want to circumvent this, you have
        #to get all the paths from somewhere.

        #######
        # all hardcoded xpaths used and attributes names:
        bravaismatrix_bulk_xpath = '/fleurInput/cell/bulkLattice/bravaisMatrix/'
        bravaismatrix_film_xpath = 'fleurInput/cell/filmLattice/bravaisMatrix/'
        species_xpath = '/fleurInput/atomSpecies/species'
        all_atom_groups_xpath = '/fleurInput/atomGroups/atomGroup'

        species_attrib_name = 'name'
        species_attrib_element = 'element'

        row1_tag_name = 'row-1'
        row2_tag_name = 'row-2'
        row3_tag_name = 'row-3'

        atom_group_attrib_species = 'species'
        atom_group_tag_abspos = 'absPos'
        atom_group_tag_relpos = 'relPos'
        atom_group_tag_filmpos = 'filmPos'
        ########

        if not ('inp.xml' in fleurinp.files):
            print('cannot get a StructureData because fleurinpdata has no inp.xml file yet')
            # TODO what to do in this case?
            return None

        # read in inpxml
        inpxmlfile = fleurinp.get_file_abs_path('inp.xml')#'./inp.xml'

        if fleurinp._schema_file_path: # Schema there, parse with schema
            xmlschema_doc = etree.parse(fleurinp._schema_file_path)
            xmlschema = etree.XMLSchema(xmlschema_doc)
            parser = etree.XMLParser(schema=xmlschema, attribute_defaults=True)
            tree = etree.parse(inpxmlfile)#, parser) # parser somewhat broken TODO, lxml version?
        else: #schema not there, parse without
            print('parsing inp.xml without XMLSchema')
            tree = etree.parse(inpxmlfile)

        root = tree.getroot()

        # Fleur uses atomic units, convert to Angstrom
        # get cell matrix from inp.xml
        row1 = root.xpath(bravaismatrix_bulk_xpath + row1_tag_name)#[0].text.split()
        cell = None

        if row1: #bulk calculation
            row1 = row1[0].text.split()
            row2 = root.xpath(bravaismatrix_bulk_xpath + row2_tag_name)[0].text.split()
            row3 = root.xpath(bravaismatrix_bulk_xpath + row3_tag_name)[0].text.split()
            # TODO? allow math?
            for i, cor in enumerate(row1):
                row1[i] = float(cor)*bohr_a
            for i, cor in enumerate(row2):
                row2[i] = float(cor)*bohr_a
            for i, cor in enumerate(row3):
                row3[i] = float(cor)*bohr_a

            cell = [row1, row2, row3]
            # create new structure Node
            struc = StructureData(cell=cell)
            struc.pbc = [True, True, True]

        elif root.xpath(bravaismatrix_film_xpath + row1_tag_name): #film calculation
            row1 = root.xpath(bravaismatrix_film_xpath + row1_tag_name)[0].text.split()
            row2 = root.xpath(bravaismatrix_film_xpath + row2_tag_name)[0].text.split()
            for i, cor in enumerate(row1):
                row1[i] = float(cor)*bohr_a
            for i, cor in enumerate(row2):
                row2[i] = float(cor)*bohr_a
            row3 = [0, 0, 0]#? TODO:what has it to be in this case?
            cell = [row1, row2, row3]
            # create new structure Node
            struc = StructureData(cell=cell)
            struc.pbc = [True, True, False]

        if cell is None:
            print('Could not extract Bravais matrix out of inp.xml. Is the '
                  'Bravais matrix explicitly given? i.e Latnam definition '
                  'not supported.')
            return None

        #get species for atom kinds
        #species = root.xpath(species_xpath)
        species_name = root.xpath(species_xpath + '/@' + species_attrib_name)
        species_element = root.xpath(species_xpath + '/@' + species_attrib_element)
        # alternativ: loop over species and species.get(species_attrib_name)

        #save species info in a dict
        species_dict = {}
        for i, spec in enumerate(species_name):
            species_dict[spec] = {species_attrib_element: species_element[i]}

        # Now we have to get all atomgroups, look what their species is and
        # their positions are.
        # Then we append them to the new structureData

        all_atom_groups = root.xpath(all_atom_groups_xpath)

        for atom_group in all_atom_groups:
            current_species = atom_group.get(atom_group_attrib_species)

            group_atom_positions_abs = atom_group.xpath(atom_group_tag_abspos)
            group_atom_positions_rel = atom_group.xpath(atom_group_tag_relpos)
            group_atom_positions_film = atom_group.xpath(atom_group_tag_filmpos)

            if group_atom_positions_abs: #we have absolut positions
                for atom in group_atom_positions_abs:
                    postion_a = atom.text.split()
                    # allow for math *, /
                    for i, pos in enumerate(postion_a):
                        if '/' in pos:
                            temppos = pos.split('/')
                            postion_a[i] = float(temppos[0])/float(temppos[1])
                        elif '*' in pos:
                            temppos = pos.split('*')
                            postion_a[i] = float(temppos[0])*float(temppos[1])
                        else:
                            postion_a[i] = float(pos)
                        postion_a[i] = postion_a[i]*bohr_a
                    # append atom to StructureData
                    struc.append_atom(
                            position=postion_a,
                            symbols=species_dict[current_species][species_attrib_element])

            elif group_atom_positions_rel: #we have relative positions
                # TODO: check if film or 1D calc, because this is not allowed! I guess
                for atom in group_atom_positions_rel:
                    postion_r = atom.text.split()
                    # allow for math * /
                    for i, pos in enumerate(postion_r):
                        if '/' in pos:
                            temppos = pos.split('/')
                            postion_r[i] = float(temppos[0])/float(temppos[1])
                        elif '*' in pos:
                            temppos = pos.split('*')
                            postion_r[i] = float(temppos[0])*float(temppos[1])
                        else:
                            postion_r[i] = float(pos)

                    # now transform to absolut Positions
                    new_abs_pos = rel_to_abs(postion_r, cell)

                    # append atom to StructureData
                    struc.append_atom(
                        position=new_abs_pos,
                        symbols=species_dict[current_species][species_attrib_element])

            elif group_atom_positions_film: # Do we support mixture always, or only in film case?
                #either support or throw error
                for atom in group_atom_positions_film:
                    # film pos are rel rel abs, therefore only transform first two coordinates
                    postion_f = atom.text.split()
                    # allow for math * /
                    for i, pos in enumerate(postion_f):
                        if '/' in pos:
                            temppos = pos.split('/')
                            postion_f[i] = float(temppos[0])/float(temppos[1])
                        elif '*' in postion_f[i]:
                            temppos = pos.split('*')
                            postion_f[i] = float(temppos[0])*float(temppos[1])
                        else:
                            postion_f[i] = float(pos)
                    # now transform to absolut Positions
                    new_abs_pos = rel_to_abs_f(postion_r, cell)
                    # append atom to StructureData
                    struc.append_atom(
                        position=new_abs_pos,
                        symbols=species_dict[current_species][species_attrib_element])
            else:
                print('I should never get here, 1D not supported yet, '
                      'I only know relPos, absPos, filmPos')
                #TODO throw error
        # TODO DATA-DATA links are not wanted, you might want to use a wf instead
        #struc.add_link_from(fleurinp, label='fleurinp.structure', link_type=LinkType.CREATE)
        #label='fleurinp.structure'
        #return {label : struc}
        return struc

    @staticmethod
    @wf
    def get_structuredata(fleurinp):
        """
        This routine return an AiiDA Structure Data type produced from the inp.xml
        file. This is a workfunction and therefore keeps the provenance.

        :return: StructureData node
        """
        return fleurinp.get_structuredata_nwf(fleurinp)



    @staticmethod
    def get_kpointsdata_nwf(fleurinp):
        """
        This routine returns an AiiDA kpoint Data type produced from the inp.xml
        file. This only works if the kpoints are listed in the in inpxml.
        This is NOT a workfunction and does not keep the provenance!
        :return: KpointsData node
        """
        from aiida.orm.data.array.kpoints import KpointsData


        #HINT, TODO:? in this routine, the 'cell' you might get in an other way
        #exp: StructureData.cell, but for this you have to make a structureData Node,
        # which might take more time for structures with lots of atoms.
        # then just parsing the cell from the inp.xml
        #as in the routine get_structureData

        #Disclaimer: this routine needs some xpath expressions.
        #these are hardcoded here, therefore maintainance might be needed,
        # if you want to circumvent this, you have
        #to get all the paths from somewhere.

        #######
        # all hardcoded xpaths used and attributes names:
        bravaismatrix_bulk_xpath = '/fleurInput/cell/bulkLattice/bravaisMatrix/'
        bravaismatrix_film_xpath = 'fleurInput/cell/filmLattice/bravaisMatrix/'
        kpointlist_xpath = '/fleurInput/calculationSetup/bzIntegration/kPointList/'

        kpoint_tag = 'kPoint'
        kpointlist_attrib_posscale = 'posScale'
        kpointlist_attrib_weightscale = 'weightScale'
        #kpointlist_attrib_count = 'count'
        kpoint_attrib_weight = 'weight'
        row1_tag_name = 'row-1'
        row2_tag_name = 'row-2'
        row3_tag_name = 'row-3'
        ########

        if not ('inp.xml' in fleurinp.files):
            print('cannot get a KpointsData because fleurinpdata has no inp.xml file yet')
            # TODO what to do in this case?
            return False

        # else read in inpxml
        inpxmlfile = fleurinp.get_file_abs_path('inp.xml')

        if fleurinp._schema_file_path: # Schema there, parse with schema
            xmlschema_doc = etree.parse(fleurinp._schema_file_path)
            xmlschema = etree.XMLSchema(xmlschema_doc)
            parser = etree.XMLParser(schema=xmlschema, attribute_defaults=True)
            tree = etree.parse(inpxmlfile, parser)
        else: #schema not there, parse without
            print('parsing inp.xml without XMLSchema')
            tree = etree.parse(inpxmlfile)

        root = tree.getroot()

        # get cell matrix from inp.xml
        cell = None
        row1 = root.xpath(bravaismatrix_bulk_xpath + row1_tag_name)#[0].text.split()

        if row1: #bulk calculation
            row1 = row1[0].text.split()
            row2 = root.xpath(bravaismatrix_bulk_xpath + row2_tag_name)[0].text.split()
            row3 = root.xpath(bravaismatrix_bulk_xpath + row3_tag_name)[0].text.split()
            # TODO? allow math?
            for i, cor in enumerate(row1):
                row1[i] = float(cor)
            for i, cor in enumerate(row2):
                row2[i] = float(cor)
            for i, cor in enumerate(row3):
                row3[i] = float(cor)

            cell = [row1, row2, row3]
            #set boundary conditions
            pbc1 = [True, True, True]

        elif root.xpath(bravaismatrix_film_xpath + row1_tag_name):
            #film calculation
            row1 = root.xpath(bravaismatrix_film_xpath + row1_tag_name)[0].text.split()
            row2 = root.xpath(bravaismatrix_film_xpath + row2_tag_name)[0].text.split()
            for i, cor in enumerate(row1):
                row1[i] = float(cor)
            for i, cor in enumerate(row2):
                row2[i] = float(cor)
            row3 = [0, 0, 0]#? TODO:what has it to be in this case?
            cell = [row1, row2, row3]
            pbc1 = [True, True, False]

        if cell is None:
            print('Could not extract Bravais matrix out of inp.xml. Is the '
                  'Bravais matrix explicitly given? i.e Latnam definition '
                  'not supported.')
            return None
        # get kpoints only works if kpointlist in inp.xml
        kpoints = root.xpath(kpointlist_xpath + kpoint_tag)

        if kpoints:
            posscale = root.xpath(kpointlist_xpath + '@' + kpointlist_attrib_posscale)
            weightscale = root.xpath(kpointlist_xpath + '@' + kpointlist_attrib_weightscale)
            #count = root.xpath(kpointlist_xpath + '@' + kpointlist_attrib_count)

            kpoints_pos = []
            kpoints_weight = []

            for kpoint in kpoints:
                kpoint_pos = kpoint.text.split()
                for i, kval in enumerate(kpoint_pos):
                    kpoint_pos[i] = float(kval)/float(posscale[0])
                    kpoint_weight = float(kpoint.get(kpoint_attrib_weight))/float(weightscale[0])
                kpoints_pos.append(kpoint_pos)
                kpoints_weight.append(kpoint_weight)
            totalw = 0
            for weight in kpoints_weight:
                totalw = totalw + weight
            kps = KpointsData(cell=cell)
            kps.pbc = pbc1

            kps.set_kpoints(kpoints_pos, cartesian=False, weights=kpoints_weight)
            #kps.add_link_from(fleurinp, label='fleurinp.kpts', link_type=LinkType.CREATE)
            kps.label='fleurinp.kpts'
            #return {label: kps}
            return kps
        else: # TODO parser other kpoints formats, if they fit in an AiiDA node
            print('No kpoint list in inp.xml')
            return None


    @staticmethod
    @wf
    def get_kpointsdata(fleurinp):
        """
        This routine returns an AiiDA kpoint Data type produced from the inp.xml
        file. This only works if the kpoints are listed in the in inpxml.
        This is a workfunction and does keep the provenance!
        :return: KpointsData node
        """

        return fleurinp.get_kpointsdata_nwf(fleurinp)

    # TODO: or move these outside...?
    #@staticmethod
    def get_parameterdata_nwf(self):
        """
        This routine returns an AiiDA ParameterData type produced from the inp.xml
        file. This node can be used for inpgen.
        This is NOT a workfunction and does NOT keep the provenance!
        :return: ParameterData node
        """
        from aiida_fleur.tools.xml_util import get_inpgen_paranode_from_xml
        if not ('inp.xml' in self.files):
            print('cannot get a StructureData because fleurinpdata has no inp.xml file yet')
            # TODO what to do in this case?
            return False

        # read in inpxml
        inpxmlfile = self.get_file_abs_path('inp.xml')#'./inp.xml'

        new_parameters = get_inpgen_paranode_from_xml(inpxmlfile)

        return new_parameters


    # Is there a way to give self to workfunctions?
    @staticmethod
    @wf
    def get_parameterdata(fleurinp):
        """
        This routine returns an AiiDA ParameterData type produced from the inp.xml
        file. This node can be used for inpgen.
        This is a workfunction and does keep the provenance!
        :return: ParameterData node
        """

        return fleurinp.get_parameterdata_nwf()





    '''
    def set_nkpts(fleurinp, count, gamma='F'):#_orgi

        kpointlist_xpath = '/fleurInput/calculationSetup/bzIntegration/kPointList'
        kpoint_xpath = '/fleurInput/calculationSetup/bzIntegration/kPoint*'
        #fleurinp = fleurinp_orgi.copy()
        if 'inp.xml' in fleurinp.files:
            # read in inpxml
            inpxmlfile = fleurinp.get_file_abs_path('inp.xml')

            if fleurinp._schema_file_path: # Schema there, parse with schema
                xmlschema_doc = etree.parse(fleurinp._schema_file_path)
                xmlschema = etree.XMLSchema(xmlschema_doc)
                parser = etree.XMLParser(schema=xmlschema, attribute_defaults=True)
                tree = etree.parse(inpxmlfile, parser)
            else: #schema not there, parse without
                print('parsing inp.xml without XMLSchema')
                tree = etree.parse(inpxmlfile)

            root = tree.getroot()
        else:
            raise InputValidationError(
                      "No inp.xml file yet specified, to add kpoints to.")

        new_kpo = etree.Element('kPointCount', count="{}".format(count), gamma="{}".format(gamma))
        print(new_kpo)
        new_tree = replace_tag(tree, kpointlist_xpath, new_kpo)
        inpxmlfile = os.path.join(
                         fleurinp._get_folder_pathsubfolder.abspath, 'temp_inp.xml')

        new_tree.write(inpxmlfile)
        print('wrote tree to' + str(inpxmlfile))
        fleurinp.del_file('inp.xml')
        fleurinp._add_path(str(inpxmlfile), 'inp.xml')
        os.remove(inpxmlfile)

        return fleurinp
    '''

    @wf
    def set_kpointsdata(fleurinp_orgi, KpointsDataNode):
        """
        This function writes the all the kpoints from a KpointsDataNode in the
        inp.xml file as a kpointslist. It replaces the Kpoints written in the
        inp.xml file.

        # currently it is the users resposibility to provide a full
        KpointsDataNode with weights. In the future FLEUR might recalculate them.
        :params: KpointsData node
        """
        from aiida.orm.data.array.kpoints import KpointsData
        #from aiida.common.exceptions import InputValidationError

        #TODO: This is probably broken and should be moved to fleurinpmodifier

        # all hardcoded xpaths used and attributes names:
        fleurinp = fleurinp_orgi.copy()
        kpointlist_xpath = '/fleurInput/calculationSetup/bzIntegration/kPointList'
        #kpoint_tag = 'kPoint'
        #kpointlist_attrib_posscale = 'posScale'
        #kpointlist_attrib_weightscale = 'weightScale'
        #kpointlist_attrib_count = 'count'
        #kpoint_attrib_weight = 'weight'

        #### method layout: #####
        # Check if Kpoints node
        # get kpoints, weights, inp.xml

        # replace the kpoints tag.(delete old write new)
        # kpoint list posScale, wightScale, count
        # <kPointList posScale="36.00000000" weightScale="324.00000000" count="324">
        #    <kPoint weight="    1.000000">   17.000000     0.000000     0.000000</kPoint>

        # add new inp.xml to fleurinpdata

        if not isinstance(KpointsDataNode, KpointsData):
            raise InputValidationError(
                      "The node given is not a valid KpointsData node.")

        if 'inp.xml' in fleurinp.files:
            # read in inpxml
            inpxmlfile = fleurinp.get_file_abs_path('inp.xml')

            if fleurinp._schema_file_path: # Schema there, parse with schema
                xmlschema_doc = etree.parse(fleurinp._schema_file_path)
                xmlschema = etree.XMLSchema(xmlschema_doc)
                parser = etree.XMLParser(schema=xmlschema, attribute_defaults=True)
                tree = etree.parse(inpxmlfile, parser)
            else: #schema not there, parse without
                print('parsing inp.xml without XMLSchema')
                tree = etree.parse(inpxmlfile)

            #root = tree.getroot()
        else:
            raise InputValidationError(
                      "No inp.xml file yet specified, to add kpoints to.")

        #cell_k = KpointsDataNode.cell

        # TODO: shall we check if cell is the same as cell from structure?
        # or is that to narrow?

        kpoint_list = KpointsDataNode.get_kpoints(also_weights=True, cartesian=False)
        nkpts = len(kpoint_list[0])
        totalw = 0
        for weight in kpoint_list[1]:
            totalw = totalw + weight
        #weightscale = totalw

        new_kpo = etree.Element('kPointList',  posScale="1.000", weightScale="1.0", count="{}".format(nkpts))
        for i, kpos in enumerate(kpoint_list[0]):
            new_k = etree.Element('kPoint', weight="{}".format(kpoint_list[1][i]))
            new_k.text = "{} {} {}".format(kpos[0], kpos[1], kpos[2])
            new_kpo.append(new_k)

        new_tree = replace_tag(tree, kpointlist_xpath, new_kpo)
        #use _write_new_fleur_xmlinp_file(fleurinp, inp_file_xmltree, fleur_change_dic):

        #TODO this should be sourced out to other methods.
        # somehow directly writing inp.xml does not work, create new one
        inpxmlfile = os.path.join(
                         fleurinp._get_folder_pathsubfolder.abspath, 'temp_inp.xml')

        # write new inp.xml, schema evaluation will be done when the file gets added
        new_tree.write(inpxmlfile)
        print('wrote tree to' + str(inpxmlfile))

        # delete old inp.xml file, not needed anymore
        #TODO maybe do some checks before
        fleurinp.del_file('inp.xml')

        # now the new inp.xml file is added and with that the inpxml_dict will
        # be overwritten, and the file validated
        fleurinp._add_path(str(inpxmlfile), 'inp.xml')

        # remove temporary inp.xml file
        os.remove(inpxmlfile)

        return fleurinp

    def get_tag(self, xpath):
        """
        Tries to evalutate an xpath expression. If it fails it logs it.

        :param root node of an etree and an xpath expression (relative, or absolute)
        :returns ALWAYS a node list
        """
        from lxml import etree

        if 'inp.xml' in self.files:
            # read in inpxml
            inpxmlfile = self.get_file_abs_path('inp.xml')

            if self._schema_file_path: # Schema there, parse with schema
                #xmlschema_doc = etree.parse(self._schema_file_path)
                #xmlschema = etree.XMLSchema(xmlschema_doc)
                #parser = etree.XMLParser(schema=xmlschema, attribute_defaults=True)
                tree = etree.parse(inpxmlfile)#, parser)
            else: #schema not there, parse without
                print('parsing inp.xml without XMLSchema')
                tree = etree.parse(inpxmlfile)

            root = tree.getroot()
        else:
            raise InputValidationError(
                      "No inp.xml file yet specified, to get a tag from")

        try:
            return_value = root.xpath(xpath)
        except etree.XPathEvalError:
            raise InputValidationError(
                'There was a XpathEvalError on the xpath: {} \n Either it does '
                'not exist, or something is wrong with the expression.'
                ''.format(xpath))
            return []
        if len(return_value) == 1:
            return return_value
        else:
            return return_value



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

"""
In this module you find the worklfow 'fleur_scf_wc' for the self-consistency
cylce management of a FLEUR calculation with AiiDA.
"""
#TODO: more info in output, log warnings
#TODO: make smarter, ggf delete broyd or restart with more or less iterations
# you can use the pattern of the density convergence for this
#TODO: other error handling, where is known what to do
#TODO: test in each step if calculation before had a problem
#TODO: maybe write dict schema for wf_parameter inputs, how?
import re
from lxml import etree
from lxml.etree import XMLSyntaxError

from aiida.orm import Code, DataFactory
from aiida.work.workchain import WorkChain, while_, if_, ToContext
from aiida.work.run import submit
from aiida.work import workfunction as wf
from aiida.common.datastructures import calc_states
from aiida_fleur.calculation.fleurinputgen import FleurinputgenCalculation
from aiida_fleur.calculation.fleur import FleurCalculation
from aiida_fleur.data.fleurinpmodifier import FleurinpModifier
from aiida_fleur.tools.common_fleur_wf import get_inputs_fleur, get_inputs_inpgen
from aiida_fleur.tools.common_fleur_wf import test_and_get_codenode, choose_resources_fleur
from aiida_fleur.tools.xml_util import eval_xpath2

RemoteData = DataFactory('remote')
StructureData = DataFactory('structure')
ParameterData = DataFactory('parameter')
FleurInpData = DataFactory('fleur.fleurinp')
FleurProcess = FleurCalculation.process()
FleurinpProcess = FleurinputgenCalculation.process()

class fleur_scf_wc(WorkChain):
    """
    Workchain for converging a FLEUR calculation (SCF).

    It converges the charge density.
    Two paths are possible:

    (1) Start from a structure and run the inpgen first optional with calc_parameters
    (2) Start from a Fleur calculation, with optional remoteData

    :param wf_parameters: (ParameterData), Workchain Spezifications
    :param structure: (StructureData), Crystal structure
    :param calc_parameters: (ParameterData), Inpgen Parameters
    :param fleurinp: (FleurinpData), to start with a Fleur calculation
    :param remote_data: (RemoteData), from a Fleur calculation
    :param inpgen: (Code)
    :param fleur: (Code)

    :return output_scf_wc_para: (ParameterData), Information of workflow results
        like Success, last result node, list with convergence behavior

    minimum input example:
    1. Code1, Code2, Structure, (Parameters), (wf_parameters)
    2. Code2, FleurinpData, (wf_parameters)

    maximum input example:
    1. Code1, Code2, Structure, Parameters
        wf_parameters: {'density_criterion' : Float,
                        'energy_criterion' : Float,
                        'converge_density' : True,
                        'converge_energy' : True}
    2. Code2, FleurinpData, (remote-data), wf_parameters as in 1.

    Hints:
    1. This workflow does not work with local codes!
    """

    _workflowversion = "0.4.0"
    _wf_default = {'fleur_runmax': 4,              # Maximum number of fleur jobs/starts (defauld 80 iterations per start)
                   'density_criterion' : 0.00002,  # Stop if charge denisty is converged below this value
                   'energy_criterion' : 0.002,     # if converge energy run also this total energy convergered below this value
                   'converge_density' : True,      # converge the charge density
                   'converge_energy' : False,      # converge the total energy (usually converged before density)
                   'refine_resources' : True,      # Tries to choose an optimal parallelisation call for a fleur calculation within the resources provieded, from the input, if set to True. 
                                                   # If False, force to use given resources and parallelisation.
                   #'resue' : True,                 # AiiDA fastforwarding (currently not there yet
                   'serial' : False,                # execute fleur with mpi or without
                   #'label' : 'fleur_scf_wc',        # label for the workchain node and all sporned calculations by the wc
                   #'description' : 'Fleur self consistensy cycle workchain', # description (see label)
                   'itmax_per_run' : 80,
                   'maxiterbroyd' : 8,
                   'inpxml_changes' : [],      # (expert) List of further changes applied after the inpgen run
                   }                                 # tuples (function_name, [parameters]), the ones from fleurinpmodifier
                                                    # example: ('set_nkpts' , {'nkpts': 500,'gamma': False}) ! no checks made, there know what you are doing
    #_default_wc_label = u'fleur_scf_wc'
    #_default_wc_description = u'fleur_scf_wc: Fleur self consistensy cycle workchain, converges the total energy.'
    _default_options = {
                        u'resources' : {"num_machines": 1},
                        u'max_wallclock_seconds' : 6*60*60,
                        u'queue_name' : u'',
                        u'custom_scheduler_commands' : u'',
                        u'import_sys_environment' : False,
                        u'environment_variables' : {}}
    
    # 0 is the usual return status of each step and the workchain
    ERROR_INVALID_INPUT_RESOURCES = 1
    ERROR_INVALID_INPUT_RESOURCES_UNDERSPECIFIED = 2
    ERROR_INVALID_CODE_PROVIDED = 3
    ERROR_INPGEN_CALCULATION_FAILED = 4
    ERROR_CHANGING_FLEURINPUT_FAILED = 5
    ERROR_CALCULATION_INVALID_INPUT_FILE = 6
    ERROR_FLEUR_CALCULATION_FALIED = 7
    ERROR_CONVERGENCE_NOT_ARCHIVED = 8
    # loop count, not enough resources provided

    @classmethod
    def define(cls, spec):
        super(fleur_scf_wc, cls).define(spec)
        spec.input("wf_parameters", valid_type=ParameterData, required=False,
                   default=ParameterData(dict=cls._wf_default))
                                              #'fleur_runmax': 4,
                                               #'density_criterion' : 0.00002,
                                               #'energy_criterion' : 0.002,
                                               #'converge_density' : True,
                                               #'converge_energy' : False,
                                               #'reuse' : True,
                                               #'options' : {
                                               #    'resources': {"num_machines": 1},
                                               #    'max_wallclock_seconds': 60*60,
                                               #    'queue_name': '',
                                               #    'custom_scheduler_commands' : '',
                                               #    #'max_memory_kb' : None,
                                               #    'import_sys_environment' : False,
                                               #    'environment_variables' : {}},
                                               #'serial' : False,
                                               #'itmax_per_run' : 30,
                                               #'inpxml_changes' : [],
                                               #}))
        spec.input("structure", valid_type=StructureData, required=False)
        spec.input("calc_parameters", valid_type=ParameterData, required=False)
        spec.input("settings", valid_type=ParameterData, required=False)
        spec.input("options", valid_type=ParameterData, required=False, 
                   default=ParameterData(dict=cls._default_options))#{
                            #'resources': {"num_machines": 1},
                            #'max_wallclock_seconds': 60*60,
                            #'queue_name': '',
                            #'custom_scheduler_commands' : '',
                            #'max_memory_kb' : None,
                            #'import_sys_environment' : False,
                            #'environment_variables' : {}}))
        spec.input("fleurinp", valid_type=FleurInpData, required=False)
        spec.input("remote_data", valid_type=RemoteData, required=False)
        spec.input("inpgen", valid_type=Code, required=False)
        spec.input("fleur", valid_type=Code, required=True)

        spec.outline(
            cls.start,
            if_(cls.validate_input)(
                cls.run_fleurinpgen),
            cls.run_fleur, # are these first runs needed TODO
            cls.inspect_fleur, # are these first runs needed
            cls.get_res, # are these first runs needed
            while_(cls.condition)(
                cls.run_fleur,
                cls.inspect_fleur,
                cls.get_res),
            cls.return_results
        )

        spec.output('fleurinp', valid_type=FleurInpData)
        spec.output('output_scf_wc_para', valid_type=ParameterData)
        spec.output('last_fleur_calc_output', valid_type=ParameterData)


        
    def start(self):
        """
        init context and some parameters
        """
        self.report(u'INFO: started convergence workflow version {}\n'
                    ''.format(self._workflowversion))

        ####### init    #######

        # internal para /control para
        self.ctx.last_calc = None
        self.ctx.loop_count = 0
        self.ctx.calcs = []
        self.ctx.abort = False


        # input para
        wf_dict = self.inputs.wf_parameters.get_dict()

        if wf_dict == {}:
            wf_dict = self._wf_default

        self.ctx.serial = wf_dict.get('serial', False)

        # set values, or defaults
        defaultoptions = self._default_options
        #options = wf_dict.get('options', defaultoptions)

        if 'options' in self.inputs:
            options = self.inputs.options.get_dict()
        else:
            options = defaultoptions
            
        for key, val in defaultoptions.iteritems():
            options[key] = options.get(key, val)
        self.ctx.options = options
        
        
        self.report(u'options: {}'.format(self.ctx.options))
        self.ctx.max_number_runs = wf_dict.get('fleur_runmax', 4)
        self.ctx.description_wf = self.inputs.get('description', '') + '|fleur_scf_wc|'
        self.ctx.label_wf = self.inputs.get('label', 'fleur_scf_wc')
        self.ctx.default_itmax = wf_dict.get('itmax_per_run', 80)
        self.ctx.refine_resources = wf_dict.get('refine_resources', True)

        # return para/vars
        self.ctx.successful = False
        self.ctx.parse_last = True
        self.ctx.distance = []
        self.ctx.total_energy = []
        self.ctx.energydiff = 10000
        self.ctx.warnings = []#
        #"warnings": {
        #"debug": {},
        #"error": {},
        #"info": {},
        #"warning": {}
        self.ctx.errors = []
        self.ctx.info = []
        self.ctx.possible_info = [
            'Consider providing more resources',
            'Consider providing a lot more resources',
            'Consider changing the mixing scheme',
            ]
        self.ctx.fleurinp = None
        self.ctx.formula = ''
        self.ctx.total_wall_time = 0
        self.ctx.last_charge_density = None
        
    def validate_input(self):
        """
        # validate input and find out which path (1, or 2) to take
        # return True means run inpgen if false run fleur directly
        """

        run_inpgen = True
        inputs = self.inputs

        if 'fleurinp' in inputs:
            run_inpgen = False
            if 'structure' in inputs:
                warning = 'WARNING: Ignoring Structure input, because Fleurinp was given'
                self.ctx.warnings.append(warning)
                self.report(warning)
            if 'inpgen' in inputs:
                warning = 'WARNING: Ignoring inpgen code input, because Fleurinp was given'
                self.ctx.warnings.append(warning)
                self.report(warning)
            if 'calc_parameters' in inputs:
                warning = 'WARNING: Ignoring parameter input, because Fleurinp was given'
                self.ctx.warnings.append(warning)
                self.report(warning)
        elif 'structure' in inputs:
            if not 'inpgen' in inputs:
                error = 'ERROR: StructureData was provided, but no inpgen code was provided'
                self.report(error)
                return self.ERROR_INVALID_INPUT_RESOURCES
        else:
            error = 'ERROR: No StructureData nor FleurinpData was provided'
            self.control_end_wc(error)
            return self.ERROR_INVALID_INPUT_RESOURCES
            
        if 'inpgen' in inputs:
            try:
                test_and_get_codenode(inputs.inpgen, 'fleur.inpgen', use_exceptions=True)
            except ValueError:
                error = ("The code you provided for inpgen of FLEUR does not "
                         "use the plugin fleur.inpgen")
                self.control_end_wc(error)
                return self.ERROR_INVALID_CODE_PROVIDED
                
        if 'fleur' in inputs:
            try:
                test_and_get_codenode(inputs.fleur, 'fleur.fleur', use_exceptions=True)
            except ValueError:
                error = ("The code you provided for FLEUR does not "
                         "use the plugin fleur.fleur")
                self.control_end_wc(error)
                return self.ERROR_INVALID_CODE_PROVIDED

        # maybe ckeck here is unessesary...
        wf_dict = self.inputs.wf_parameters.get_dict()

        if wf_dict == {}:
            wf_dict = self._wf_default

        # check format of inpxml_changes
        fchanges = wf_dict.get('inpxml_changes', [])
        if fchanges:
            for change in fchanges:
                # somehow the tuple type gets destroyed on the way and becomes a list
                if (not isinstance(change, tuple)) and (not isinstance(change, list)):
                    error = ('ERROR: Wrong Input inpxml_changes wrong format of'
                             ': {} should be tuple of 2. I abort'.format(change))
                    self.control_end_wc(error)
                    self.ERROR_INVALID_INPUT_RESOURCES

        return run_inpgen


    def run_fleurinpgen(self):
        """
        run the inpgen
        """
        structure = self.inputs.structure
        self.ctx.formula = structure.get_formula()
        self.ctx.natoms = len(structure.sites)
        label = 'scf: inpgen'
        description = '{} inpgen on {}'.format(self.ctx.description_wf, self.ctx.formula)

        inpgencode = self.inputs.inpgen
        if 'calc_parameters' in self.inputs:
            params = self.inputs.calc_parameters
        else:
            params = None
        
        options = {"max_wallclock_seconds" : int(self.ctx.options.get('max_wallclock_seconds')),
                   "resources" : self.ctx.options.get('resources', {"num_machines": 1}),
                   "queue_name" : self.ctx.options.get('queue_name', '')}
        # TODO do not use the same option for inpgen as for FLEUR... so far we ignore the others...
        # clean Idea might be to provide second inpgen options, currenly for our purposes not nessesary...
        self.report(options)
        inputs_build = get_inputs_inpgen(structure, inpgencode, options, label, description, params=params)
        self.report('INFO: run inpgen')
        future = submit(FleurinpProcess, **inputs_build)

        return ToContext(inpgen=future, last_calc=future)

    def change_fleurinp(self):
        """
        This routine sets somethings in the fleurinp file before running a fleur
        calculation.
        """
        self.report(u'INFO: run change_fleurinp')
        if self.ctx.fleurinp: #something was already changed
            #print('Fleurinp already exists')
            return 
        elif 'fleurinp' in self.inputs:
            fleurin = self.inputs.fleurinp
        else:
            try:
                fleurin = self.ctx['inpgen'].out.fleurinpData
            except AttributeError:
                error = 'No fleurinpData found, inpgen failed'
                self.control_end_wc(error)
                return self.ERROR_INPGEN_CALCULATION_FAILED

        wf_dict = self.inputs.wf_parameters.get_dict()
        converge_te = wf_dict.get('converge_energy', False)
        fchanges = wf_dict.get('inpxml_changes', [])

        if not converge_te or fchanges:# change inp.xml file
            fleurmode = FleurinpModifier(fleurin)
            if not converge_te:
                dist = wf_dict.get('density_criterion', 0.00002)
                maxbroyd = wf_dict.get('maxiterbroyd', 8)
                fleurmode.set_inpchanges({'itmax': self.ctx.default_itmax, 'minDistance' : dist, 'maxIterBroyd' : maxbroyd})
            avail_ac_dict = fleurmode.get_avail_actions()

            # apply further user dependend changes
            if fchanges:
                for change in fchanges:
                    function = change[0]
                    para = change[1]
                    method = avail_ac_dict.get(function, None)
                    if not method:
                        error = ("ERROR: Input 'inpxml_changes', function {} "
                                 "is not known to fleurinpmodifier class, "
                                 "plaese check/test your input. I abort..."
                                 "".format(method))
                        self.control_end_wc(error)
                        return self.ERROR_CHANGING_FLEURINPUT_FAILED

                    else:# apply change
                        method(**para)

            # validate?
            apply_c = True
            try:
                fleurmode.show(display=False, validate=True)
            except XMLSyntaxError:
                error = ('ERROR: input, user wanted inp.xml changes did not validate')
                #fleurmode.show(display=True)#, validate=True)
                self.report(error)
                apply_c = False
                return self.ERROR_CALCULATION_INVALID_INPUT_FILE
            
            # apply
            if apply_c:
                out = fleurmode.freeze()
                self.ctx.fleurinp = out
            return
        else: # otherwise do not change the inp.xml
            self.ctx.fleurinp = fleurin
            return


    def run_fleur(self):
        """
        run a FLEUR calculation
        """
        self.report(u'INFO: run FLEUR')

        self.change_fleurinp()
        fleurin = self.ctx.fleurinp
        
        # TODO: ggf move the resource refinement somewhere else should be done only once, and maybe again if some parameters were changed in fleurinp
        if self.ctx.refine_resources:
            options = self.ctx.options.copy()
            # get_kpts # TODO: try to do this more efficient... also this can fail...
            kpt = fleurin.get_kpointsdata_nwf(fleurin)
            kpt_shape = kpt.get_shape('kpoints')
            nkpt = kpt_shape[0]
            # TODO; get computer memmory and ncores per node from computer and scheduler
            if not self.ctx.natoms:
                structure = fleurin.get_structuredata_nwf(fleurin)
                self.ctx.natoms = len(structure.sites)
            resources = options.get(u'resources', {u"num_machines": 1})
            ncores_per_node = 24 # TODO
            memmory_gb = 120 # TODO
            optimal_res = choose_resources_fleur(nkpt=nkpt, natm=self.ctx.natoms, max_resources=resources, ncores_per_node=ncores_per_node, memmory_gb=memmory_gb)
            #optimal_res: nodes, mpi_per node, openmp_num_thread, warnings
            
            option_keys = options.keys()
            res_keys = resources.keys()
            if 'tot_num_mpiprocs' in res_keys:
                new_resources = {'tot_num_mpiprocs' : optimal_res[0]*optimal_res[1]}
                if 'num_mpiprocs_per_machine' in res_keys:
                    new_resources['num_mpiprocs_per_machine'] = optimal_res[1]
                options['resources'] = new_resources
            elif 'num_machines' in res_keys:
                new_resources = {'num_machines' : optimal_res[0]}
                options[u'resources'] = new_resources
            
            if u'custom_scheduler_commands' in option_keys:
                sched_comd = options[u'custom_scheduler_commands']
                if 'span[ptile=' in sched_comd:
                    sched_comd = re.sub(u'ptile=\d?\d', u'ptile={}'.format(optimal_res[1]), sched_comd)
                options[u'custom_scheduler_commands'] = sched_comd
                
            if u'environment_variables' in option_keys:
                environment_variables = options[u'environment_variables']
                environment_variables[u'OMP_NUM_THREADS'] = u'{}'.format(optimal_res[2])
                options[u'environment_variables'] = environment_variables
            
            if u'max_memory_kb' in option_keys:
                max_memory_kb = int(memmory_gb*1000/optimal_res[1])
                options[u'max_memory_kb'] = max_memory_kb
            
            self.ctx.options = options
            self.report(u'changed options to {}'.format(options))
        else:
            options = self.ctx.options.copy()
        
        '''
        if 'settings' in self.inputs:
            settings = self.input.settings
        else:
            settings = ParameterData(dict={'files_to_retrieve' : [],
                                           'files_not_to_retrieve': [],
                                           'files_copy_remotely': [],
                                           'files_not_copy_remotely': [],
                                           'commandline_options': ["-wtime", "{}".format(self.ctx.options['max_wallclock_seconds'])],
                                           'blaha' : ['bla']})
        '''
        if self.ctx['last_calc']:
            # will this fail if fleur before failed? try needed?
            remote = self.ctx['last_calc'].out.remote_folder
        elif u'remote_data' in self.inputs:
            remote = self.inputs.remote_data
        else:
            remote = None

        label = ' '
        description = ' '
        if self.ctx.formula:
            label = u'scf: fleur run {}'.format(self.ctx.loop_count+1)
            description = u'{} fleur run {} on {}'.format(self.ctx.description_wf, self.ctx.loop_count+1, self.ctx.formula)
        else:
            label = u'scf: fleur run {}'.format(self.ctx.loop_count+1)
            description = u'{} fleur run {}, fleurinp given'.format(self.ctx.description_wf, self.ctx.loop_count+1)

        code = self.inputs.fleur


        
        inputs_builder = get_inputs_fleur(code, remote, fleurin, options, label, description, serial=self.ctx.serial)
        future = submit(FleurProcess, **inputs_builder)
        self.ctx.loop_count = self.ctx.loop_count + 1
        self.report(u'INFO: run FLEUR number: {}'.format(self.ctx.loop_count))
        self.ctx.calcs.append(future)

        return ToContext(last_calc=future)

    def inspect_fleur(self):
        """
        Analyse the results of the previous Calculation (Fleur or inpgen),
        checking whether it finished successfully or if not, troubleshoot the
        cause and adapt the input parameters accordingly before
        restarting, or abort if unrecoverable error was found
        """
        #expected_states = [calc_states.FINISHED, calc_states.FAILED,
        #                   calc_states.SUBMISSIONFAILED]
        #print(self.ctx['last_calc'])
        #self.report(u'I am in inspect_fleur')
        try:
            calculation = self.ctx.last_calc
        except AttributeError:
            self.ctx.parse_last = False
            error = 'ERROR: Something went wrong I do not have a last calculation'
            self.control_end_wc(error)
            return self.ERROR_FLEUR_CALCULATION_FALIED
        calc_state = calculation.get_state()
        #self.report(u'the state of the last calculation is: {}'.format(calc_state))

        if calc_state != calc_states.FINISHED:
            error = ('ERROR: Last Fleur calculation failed somehow it is '
                    'in state {}'.format(calc_state))
            self.control_end_wc(error)
            return self.ERROR_FLEUR_CALCULATION_FALIED
        elif calc_state == calc_states.FINISHED:
            self.ctx.parse_last = True

        '''
        # Done: successful convergence of last calculation
        if calculation.has_finished_ok():
            self.report(u'converged successfully after {} iterations'.format(self.ctx.iteration))
            self.ctx.restart_calc = calculation
            self.ctx.is_finished = True

        # Abort: exceeded maximum number of retries
        elif self.ctx.iteration >= self.ctx.max_iterations:
            self.report(u'reached the max number of iterations {}'.format(self.ctx.max_iterations))
            self.abort_nowait('last ran PwCalculation<{}>'.format(calculation.pk))

        # Abort: unexpected state of last calculation
        elif calculation.get_state() not in expected_states:
            self.abort_nowait('unexpected state ({}) of PwCalculation<{}>'.format(
                calculation.get_state(), calculation.pk))

        # Retry: submission failed, try to restart or abort
        elif calculation.get_state() in [calc_states.SUBMISSIONFAILED]:
            self._handle_submission_failure(calculation)

        # Retry: calculation failed, try to salvage or abort
        elif calculation.get_state() in [calc_states.FAILED]:
            self._handle_calculation_failure(calculation)

        # Retry: try to convergence restarting from this calculation
        else:
            self.report(u'calculation did not converge after {} iterations, restarting'.format(self.ctx.iteration))
            self.ctx.restart_calc = calculation

        # ggf handle bad convergence behavior, delete broyd (do not copy)
        return
        '''

    def get_res(self):
        """
        Check how the last Fleur calculation went
        Parse some results.
        """
        # TODO maybe do this different
        # or if complexer output node exists take from there.

        xpath_energy = '/fleurOutput/scfLoop/iteration/totalEnergy/@value'
        xpath_distance = '/fleurOutput/scfLoop/iteration/densityConvergence/chargeDensity/@distance' # be aware of magnetism
        #densityconvergence_xpath = 'densityConvergence'
        #chargedensity_xpath = 'densityConvergence/chargeDensity'
        #overallchargedensity_xpath = 'densityConvergence/overallChargeDensity'
        #spindensity_xpath = 'densityConvergence/spinDensity'
        if self.ctx.parse_last:#self.ctx.successful:
            #self.report('last calc successful = {}'.format(self.ctx.successful))
            last_calc = self.ctx.last_calc

            '''
            spin = get_xml_attribute(eval_xpath(root, magnetism_xpath), jspin_name)

                charge_densitys = eval_xpath(iteration_node, chargedensity_xpath)
                charge_density1 = get_xml_attribute(charge_densitys[0], distance_name)
                write_simple_outnode(
                    charge_density1, 'float', 'charge_density1', simple_data)

                charge_density2 = get_xml_attribute(charge_densitys[1], distance_name)
                write_simple_outnode(
                    charge_density2, 'float', 'charge_density2', simple_data)

                spin_density = get_xml_attribute(
                    eval_xpath(iteration_node, spindensity_xpath), distance_name)
                write_simple_outnode(
                    spin_density, 'float', 'spin_density', simple_data)

                overall_charge_density = get_xml_attribute(
                    eval_xpath(iteration_node, overallchargedensity_xpath), distance_name)
                write_simple_outnode(
                    overall_charge_density, 'float', 'overall_charge_density', simple_data)

            '''
            #TODO: dangerous, can fail, error catching
            # TODO: is there a way to use a standard parser?
            outxmlfile = last_calc.out.output_parameters.dict.outputfile_path
            walltime = last_calc.out.output_parameters.dict.walltime
            if isinstance(walltime, int):
                self.ctx.total_wall_time = self.ctx.total_wall_time + walltime
            #outpara = last_calc.get('output_parameters', None)
            #outxmlfile = outpara.dict.outputfile_path
            parser = etree.XMLParser(recover=False)#, remove_blank_text=True)
            outfile_whole = True
            try:
                tree = etree.parse(outxmlfile, parser)
            except etree.XMLSyntaxError:
                outfile_whole = False
                errormsg = ('ERROR: SCF wc was not successful. The out.xml file of the last Fleur calculation is broken.'
                           ' FLEUR calc failed, through it was marked as FINISHED... I abort.')
                self.control_end_wc(errormsg)
                return self.ERROR_FLEUR_CALCULATION_FALIED
            
            if outfile_whole:
                root = tree.getroot()
                energies = eval_xpath2(root, xpath_energy)
                for energy in energies:
                    self.ctx.total_energy.append(float(energy))

                distances = eval_xpath2(root, xpath_distance)
                for distance in distances:
                    self.ctx.distance.append(float(distance))
        else:
            errormsg = 'ERROR: scf wc was not successful, check log for details'
            self.control_end_wc(errormsg)
            return self.ERROR_FLEUR_CALCULATION_FALIED
            # otherwise this will lead to erros further down

    def condition(self):
        """
        check convergence condition
        """

        density_converged = False
        energy_converged = False
        # TODO do a test first if last_calculation was successful, otherwise,
        # 'output_parameters' wont exist.
        inpwfp_dict = self.inputs.wf_parameters.get_dict()
        #last_charge_density = self.ctx.last_calc['output_parameters'].dict.charge_density
        # not a good fix for magnetic stuff, but for now, we want to test if the rest works.
        try:
            last_charge_density = self.ctx.last_calc.out.output_parameters.dict.charge_density
        except AttributeError:
            # magnetic system
            last_charge_density = self.ctx.last_calc.out.output_parameters.dict.overall_charge_density
            # divide by 2?
        self.ctx.last_charge_density = last_charge_density
        if inpwfp_dict.get('converge_density', True):
            if inpwfp_dict.get('density_criterion', 0.00002) >= last_charge_density:
                density_converged = True
        else:
            density_converged = True #since density convergence is not wanted

        energy = self.ctx.total_energy

        if len(energy) >= 2:
            self.ctx.energydiff = abs(energy[-1]-energy[-2])
        if inpwfp_dict.get('converge_energy', True):
            if inpwfp_dict.get('energy_criterion', 0.002) >= self.ctx.energydiff:
                energy_converged = True
        else:
            energy_converged = True #since energy convergence is not wanted

        if density_converged and energy_converged:
            self.ctx.successful = True
            return False
        elif self.ctx.loop_count >= self.ctx.max_number_runs:
            self.ctx.successful = False
            return False
        else:
            return True


    def return_results(self):
        """
        return the results of the calculations
        This shoudl run through and produce output nodes even if everything failed,
        therefore it only uses results from context.
        """
        try:
            last_calc_uuid = self.ctx.last_calc.uuid
        except AttributeError:
            last_calc_uuid = None
        try: # if something failed, we still might be able to retrieve something
            last_calc_out = self.ctx.last_calc.out['output_parameters']
            retrieved = self.ctx.last_calc.out['retrieved']
            last_calc_out_dict = last_calc_out.get_dict()
        except KeyError:
            last_calc_out = None
            last_calc_out_dict = {}
            retrieved = None



        outputnode_dict = {}
        outputnode_dict['workflow_name'] = self.__class__.__name__
        outputnode_dict['workflow_version'] = self._workflowversion
        outputnode_dict['material'] = self.ctx.formula
        outputnode_dict['loop_count'] = self.ctx.loop_count
        outputnode_dict['iterations_total'] = last_calc_out_dict.get('number_of_iterations_total', None)
        outputnode_dict['distance_charge'] = self.ctx.last_charge_density #last_calc_out_dict.get('charge_density', None)
        outputnode_dict['distance_charge_all'] = self.ctx.distance
        outputnode_dict['total_energy'] = last_calc_out_dict.get('energy_hartree', None)
        outputnode_dict['total_energy_all'] = self.ctx.total_energy
        outputnode_dict['distance_charge_units'] = 'me/bohr^3'
        outputnode_dict['total_energy_units'] = 'Htr'
        outputnode_dict['last_difference_total_energy'] = self.ctx.energydiff
        outputnode_dict['last_difference_total_energy_units'] = 'Htr'
        outputnode_dict['successful'] = self.ctx.successful
        outputnode_dict['last_calc_uuid'] = last_calc_uuid
        outputnode_dict['total_wall_time'] = self.ctx.total_wall_time
        outputnode_dict['total_wall_time_units'] = 'seconds'
        outputnode_dict['info'] = self.ctx.info
        outputnode_dict['warnings'] = self.ctx.warnings
        outputnode_dict['errors'] = self.ctx.errors

        # maybe also store some information about the formula
        #also lognotes, which then can be parsed from subworkflow too workflow, list of calculations involved (pks, and uuids),
        #This node should contain everything you wish to plot, here iteration versus, total energy and distance.

        if self.ctx.successful:
            self.report(u'STATUS: Done, the convergence criteria are reached.\n'
                        'INFO: The charge density of the FLEUR calculation '
                        'converged after {} FLEUR runs, {} iterations and {} sec '
                        'walltime to {} "me/bohr^3" \n'
                        'INFO: The total energy difference of the last two iterations '
                        'is {} htr \n'.format(self.ctx.loop_count,
                                       last_calc_out_dict.get('number_of_iterations_total', None),
                                       self.ctx.total_wall_time,
                                       last_calc_out_dict.get('charge_density', None), 
                                       self.ctx.energydiff))

        else: # Termination ok, but not converged yet...
            if self.ctx.abort: # some error occured, donot use the output.
                self.report(u'STATUS/ERROR: I abort, see logs and '
                            'erros/warning/hints in output_scf_wc_para')
            else:
                self.report(u'STATUS/WARNING: Done, the maximum number of runs '
                            'was reached or something failed.\n INFO: The '
                            'charge density of the FLEUR calculation, '
                            'after {} FLEUR runs, {} iterations and {} sec '
                            'walltime is {} "me/bohr^3"\n'
                            'INFO: The total energy difference of the last '
                            'two interations is {} htr'
                            ''.format(self.ctx.loop_count,
                            last_calc_out_dict.get('number_of_iterations_total', None),
                            self.ctx.total_wall_time,
                            last_calc_out_dict.get('charge_density', None), self.ctx.energydiff))

        #also lognotes, which then can be parsed from subworkflow too workflow, list of calculations involved (pks, and uuids),
        #This node should contain everything you wish to plot, here iteration versus, total energy and distance.


        outputnode_t = ParameterData(dict=outputnode_dict)
         # this is unsafe so far, because last_calc_out could not exist...
        if last_calc_out:
            outdict = create_scf_result_node(outpara=outputnode_t, last_calc_out=last_calc_out, last_calc_retrieved=retrieved)
        else:
            outdict = create_scf_result_node(outpara=outputnode_t)

        if 'fleurinp' in self.inputs:
            outdict['fleurinp'] = self.inputs.fleurinp
        else:
            try:
                fleurinp = self.ctx['inpgen'].out.fleurinpData
            except AttributeError:
                self.report(u'ERROR: No fleurinp, something was wrong with the inpgen calc')
                fleurinp = None
                self.control_end_wc(u'ERROR: No fleurinp, something was wrong with the inpgen calc') # in the aiida 0.12 version
            outdict['fleurinp'] = fleurinp
        if last_calc_out:
            outdict[u'last_fleur_calc_output'] = last_calc_out

        #outdict['output_scf_wc_para'] = outputnode
        for link_name, node in outdict.iteritems():
            self.out(link_name, node)



    def handle_fleur_failure(self):
        """
        handle a failure of a Fleur calculation
        """
        return
        # handle out of walltime (not one interation run) abort, tell user to specifi mor resources, or different cutoffs

        # handle fleur error fermi level convergence
        # restart fleur with more tempertature broad or other smearing type (gauss = T)

        # qfix needed, restart fleur with Qfix

        # differ errors, solving dirac equation
        # abort, or restart with different parameters

        # muffin tin radii overlap, restart with smaller MT

        # wrong command line switches
        # hdf, -h , ... restart with the right one, or abort

        # something about kpt grid
        # abort, or restart with increased number /different number of kpoints

        # wrong parallelisation
        # abort or restart with right parallelisation

        #

        '''
        #ALl FLEUR current (07.2017) ERROR hints: TODO

        1.nsh greater than dimensioned | increase nmax in jcoff2'
        2. Fermi-level determination did not converge| change temperature or set input = F
        3. Too low eigenvalue detected | If the lowest eigenvalue is more than 1Htr below the lowest energy parameter, you probably have picked up a ghoststate
        4. e >= vz0 | Vacuum energy-parameter too high
        5. kidx /= stars | something wrong with kmxxc_fft or nxc3_fft
        6 Wrong size of matrix | Your basis might be too large or the parallelization fail or ??
        If no LO's are set skiplo must be 0 in enpara
        You compiled without support for HDF5. Please use another mode
        Your HDF5 library does not support parallel IO
        Use -mpi or -hdf instead
        Consider setting 'autocomp = false' in apwefl.
        Film setup not centered" | The z = 0 plane is the center of the film
        setcore_bystr |
        expnadconfig |
        File not readable:"//filename | FLEUR wants to read a file that is not present
        Both inp & inp.xml given  | Please delete one of the input files or specify -xml to use inp.xml
        inp.xml not found | You gave the -xml option but provided no inp.xml file
        No input file found | To use FLEUR, you have to provide either an 'inp' or an 'inp.xml' file in the working directory
        Use a supercell or a different functional
        MPI parallelization failed | ou have to either compile FLEUR with a parallel diagonalization library (ELPA,SCALAPACK...) or you have to run such that the No of kpoints can be distributed on the PEs

        AiiDA related:
        1. Submissionfailed:
        SSHException: SSH session not active   -> wait and relaunch

        Scheduler/HPC related:

        1. scf scheduler needs other things than torque, try to provide both as default
        that they both work

        2. ran out of walltime with out a single interation done. (killed by scheduler)
        -> terminate, because the user should specify more recources or in the
        calculation something is fishy

        3. bumped by scheduler joblimit. reached
        -> think about something here and retry

        4. no space/files left on remote machine.
        -> also terminate, tell user to delete stuff

        5. not eneough memory, (killed by scheduler, or some fleur error)
        -> can we do something here
        '''


    def handle_inpgen_failure(self):
        """
        Handle a failure of inpgen
        """
        return

    def control_end_wc(self, errormsg):
        """
        Controled way to shutdown the workchain. will initalize the output nodes
        The shutdown of the workchain will has to be done afterwards
        """
        self.ctx.successful = False
        self.ctx.abort = True
        self.report(errormsg) # because return_results still fails somewhen
        self.ctx.errors.append(errormsg)
        self.return_results()
        
        return
'''
if __name__ == "__main__":
    import argparse
    from aiida.orm import load_node
    from aiida_fleur.tools.common_fleur_wf import is_code
    
    parser = argparse.ArgumentParser(description=('SCF with FLEUR. workflow to'
                 ' converge the chargedensity and optional the total energy. all arguments are pks, or uuids, codes can be names'))
    parser.add_argument('--wf_para', type=ParameterData, dest='wf_parameters',
                        help='Some workflow parameters', required=False)
    parser.add_argument('--structure', type=StructureData, dest='structure',
                        help='The crystal structure node', required=False)
    parser.add_argument('--calc_para', type=ParameterData, dest='calc_parameters',
                        help='Parameters for the FLEUR calculation', required=False)
    parser.add_argument('--fleurinp', type=FleurInpData, dest='fleurinp',
                        help='FleurinpData from which to run the FLEUR calculation', required=False)
    parser.add_argument('--remote', type=RemoteData, dest='remote_data',
                        help=('Remote Data of older FLEUR calculation, '
                        'from which files will be copied (broyd ...)'), required=False)
    parser.add_argument('--inpgen', type=Code, dest='inpgen',
                        help='The inpgen code node to use', required=False)
    parser.add_argument('--fleur', type=Code, dest='fleur',
                        help='The FLEUR code node to use', required=True)

    args = parser.parse_args()
    
    # load_the nodes
    #if args.wf_parameters:
    wf_parameters = load_node(args.wf_parameters)
    
    structure = load_node(args.structure)
    
    #if args.calc_parameters:
    calc_parameters = load_node(args.calc_parameters)
    
    fleurinp = load_node(args.fleurinp)
    remote_data = load_node(args.remote_data)
    
    inpgen = is_code(args.inpgen)       
    fleur = is_code(args.fleur)    
    
    # TODO input logic....
    
    # submit fleur_scf_wc with different inputs.
    
    #res = submit(fleur_scf_wc,
    #          wf_parameters=wf_parameters,
    #          structure=structure,
    #          calc_parameters=args.calc_parameters,
    #          fleurinp=args.fleurinp,
    #          remote_data=args.remote_data,
    #          inpgen = args.inpgen,
    #          fleur=args.fleur)
'''


@wf
def create_scf_result_node(**kwargs):
    """
    This is a pseudo wf, to create the rigth graph structure of AiiDA.
    This wokfunction will create the output node in the database.
    It also connects the output_node to all nodes the information commes from.
    So far it is just also parsed in as argument, because so far we are to lazy
    to put most of the code overworked from return_results in here.
    """
    for key, val in kwargs.iteritems():
        if key == 'outpara': #  should be alwasys there
            outpara = val
    outdict = {}
    outputnode = outpara.copy()#clone()
    outputnode.label = 'output_scf_wc_para'
    outputnode.description = ('Contains self-consistency results and '
                             'information of an fleur_scf_wc run.')

    outdict['output_scf_wc_para'] = outputnode
    # copy, because we rather produce the same node twice then have a circle in the database for now...
    #output_para = args[0]
    #return {'output_eos_wc_para'}
    return outdict






# write_results_to_file
def test_write_results_to_file_interface():
    """
    is the basic file writter wraper working as indended
    """
    from aiida_fleur.tools.io_routines import write_results_to_file
    from os.path import isfile, abspath
    from os import remove
    #import os
    from numpy import array
    

    # testing some defaults
    inputhead = 'head\n'
    data = array([[1,2],[3,4]])
    destination = './outputfiletest'
    write_results_to_file(inputhead, data, destination=destination)
    isfile_ = isfile(abspath('./outputfiletest'))
    test_file = open(destination, 'r')
    content = test_file.read()
    test_file.close()

    content_exp = 'head\n1.00000000  3.00000000\n2.00000000  4.00000000\n'
    remove(destination)

    assert isfile_
    assert content == content_exp


# write_xps_spectra_datafile
def test_write_xps_spectra_datafile_interface():
    """
    is the xps data file writter working, is the file ok?
    """
    from aiida_fleur.tools.io_routines import write_xps_spectra_datafile
    
    #TODO how to test this?
    # provide all sample inputs and check contents of outputfile
    pass
    #assert 1 == 2



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the AiiDA-FLEUR package.                               #
#                                                                             #
# The code is hosted on GitHub at https://github.com/broeder-j/aiida-fleur    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
# http://aiida-fleur.readthedocs.io/en/develop/                               #
###############################################################################

# Here we test if the interfaces of the workflows are still the same
import pytest
'''
@pytest.mark.usefixtures("aiida_env")
class TestFleur_workflow_interfaces:#TestAiida_fleur_entrypoints
    """
    Test all aiida-fleur workflow interfaces
    """
    # TODO
    # prepare some nodes:
    # structure, option, fleurinp, wfparameters
    # add to builder and see if he takes it
    # ggf if possible run initial step only, that the input is checked...
    
    def test_fleur_scf_wc_init(aiida_env):
        """
        Test the interface of the scf workchain
        """
        from aiida_fleur.workflows.scf import fleur_scf_wc
        
        builder = fleur_scf_wc.get_builder()

        
    def test_fleur_eos_wc_init(aiida_env):
        """
        Test the interface of the eos workchain
        """
        from aiida_fleur.workflows.eos import fleur_eos_wc
        
        builder = fleur_eos_wc.get_builder()


    def test_fleur_dos_wc_init(aiida_env):
        """
        Test the interface of the dos workchain
        """
        from aiida_fleur.workflows.dos import fleur_dos_wc
        
        builder = fleur_dos_wc.get_builder()


    def test_fleur_band_wc_init(aiida_env):
        """
        Test the interface of the band workchain
        """
        from aiida_fleur.workflows.scf import fleur_scf_wc
        
        builder = fleur_scf_wc.get_builder()


    #def test_fleur_band2_wc_init(aiida_env):
    #    """
    #    Test the interface of the band2 workchain
    #    """
    #    from aiida_fleur.workflows.band2 import fleur_band2_wc
    #    
    #    builder = fleur_band2_wc.get_builder()


    def test_fleur_corehole_wc_init(aiida_env):
        """
        Test the interface of the corehole workchain
        """
        from aiida_fleur.workflows.corehole import fleur_corehole_wc
        
        builder = fleur_corehole_wc.get_builder()


    def test_fleur_initial_cls_wc_init(aiida_env):
        """
        Test the interface of the scf workchain
        """
        from aiida_fleur.workflows.initial_cls import fleur_initial_cls_wc
        
        builder = fleur_initial_cls_wc.get_builder()


    def test_fleur_delta_wc_init(aiida_env):
        """
        Test the interface of the delta workchain
        """
        from aiida_fleur.workflows.delta import fleur_delta_wc
        
        builder = fleur_delta_wc.get_builder()


    def test_fleur_relax_wc_init(aiida_env):
        """
        Test the interface of the relax workchain
        """
        from aiida_fleur.workflows.relax import fleur_relax_wc
        
        builder = fleur_relax_wc.get_builder()


    def test_fleur_optimize_para_wc_init(aiida_env):
        """
        Test the interface of the optimize_para_ workchain
        """
        from aiida_fleur.workflows.optimize_para import fleur_optimize_parameters_wc
        
        builder = fleur_optimize_parameters_wc.get_builder()

'''
